#!/usr/bin/env python3

import parglare as Parglare

class Grammar:
    def __init__(self, grammar, keywords, recognizers=None, debug=False):
        assert isinstance(grammar, str)
        self.grammar = grammar
        self.keywords = keywords
        self.recognizers = recognizers
        self.debug = debug

    def get_action(self,inp):
        raise Exception("override this")
    
    def interpret_result(self, result):
        # Do nothing
        return result

    def _get_parser(self, actions={}):
        if self.recognizers is not None:
            g = Parglare.Grammar.from_string(self.grammar, recognizers=self.recognizers, debug=self.debug)
        else:
            g = Parglare.Grammar.from_string(self.grammar)
        parser = Parglare.Parser(g, debug=self.debug, actions=actions)
        return parser

    def compile(self, code):
        parser = self._get_parser()
        result = parser.parse(code)
        return result

    def run(self, code, inp):
        actions = self.get_action(inp)
        parser = self._get_parser(actions)
        result = parser.parse(code)
        return self.interpret_result(result)
    
    def get_syntax_tree(self, code):
        g = Parglare.Grammar.from_string(self.grammar)
        parser = Parglare.Parser(g, debug=False, build_tree=True)
        tree = parser.parse(code)
        return tree

    def get_success_program_idx(self, programs):
        for idx, soln in enumerate(programs):
            try:
                examples = soln["Examples"]
                code = soln["Program"]
                all_example_match = False
                for ex in examples:
                    inp = ex["In"]
                    expected_out = ex["Out"]
                    actual_out = self.run(code, inp)
                    if str(actual_out) == str(expected_out):
                        all_example_match = True
                    else:
                        break
                if all_example_match:
                    yield idx
            except:
                pass
    
    def get_success_programs_unique(self, programs):
        prog_set = set()
        for idx in self.get_success_program_idx(programs):
            prog = programs[idx]
            if prog["Program"] not in prog_set:
                prog_set.add(prog["Program"])
                yield prog
        prog_set.clear()
    
    def check_preconditions(self, precondition: str):
        return "[]"

    def check_postconditions(self, postcondition: str, inp: str, output: str):
        pass


#!/usr/bin/env python3

import sys

root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
import os
import json
import ray
import omegaconf
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from enum import Enum
from itp_interface.rl.proof_tree import ProofSearchResult
from itp_interface.rl.proof_action import ProofAction
from itp_interface.rl.simple_proof_env import ProofEnvReRankStrategy
from proof_wala.search.search import SearchAlgorithm
from proof_wala.proof_search.search_driver import ProofPathTracer
from proof_wala.search.beam_search import BeamSearch
from proof_wala.search.best_first_search import BestFirstSearch
from proof_wala.proof_search.llm_tactic_generator import NegLoglikelihoodMinimizingHeuristic

class SettingType(Enum):
    Agent = "Agent"
    GptF = "GptF"

    def __str__(self):
        return self.value

@dataclass_json
@dataclass
class EnvSettings(object):
    name: str
    retrieval_strategy: ProofEnvReRankStrategy

@dataclass_json
@dataclass
class EvalSettings(object):
    name: str
    use_hammer: bool
    model_name: str
    is_seq2seq: bool
    max_proof_depth: int = 50
    max_seq_length: int = 2048
    timeout_in_secs: int = 60
    proof_retries: int = 1
    character_per_token: float = 3.6
    width: int = 64
    max_tokens_per_action: int = 175
    do_sample: bool = True
    top_k: int = 20
    stop_tokens: typing.List[str] = field(default_factory=lambda: ["[END]"])
    padding: bool = True
    return_full_text: bool = False
    compute_probabilities: bool = True
    checkpoint_dir: str = ".log/checkpoints"
    should_checkpoint: bool = False
    temperature: float = 0.0
    max_history_messages: int = 0
    proof_dump_dir: str = ".log/proofs/proof-dump-"
    use_human_readable_proof_context: bool = True
    sample: float = 1.0
    sample_seed: int = 0xf00
    use_example_retrieval: bool = False
    always_use_useful_theorem_retrieval: bool = False
    num_goal_per_prompt: typing.Optional[int] = None
    search_strategy: str = "BeamSearch"
    search_params: typing.Dict[str, typing.Any] = field(default_factory=lambda: {"width": 64, "max_new_tokens": 175, "temperature": 0.75, "do_sample": True, "top_k": 64, "padding": True, "return_full_text": False, "compute_probabilities": True})
    proof_search_heuristic: str = "NegLoglikelihoodMinimizingHeuristic"
    model_parallelism: int = 4
    do_lemmas_discovery: bool = True
    proof_tracer: typing.Optional[ProofPathTracer] = None

    @property
    def should_capture_traces_while_proof_search(self):
        return self.proof_tracer is not None

    def get_search_algo(self):
        if self.search_strategy == "BeamSearch":
            return BeamSearch(**self.search_params)
        elif self.search_strategy == "BestFirstSearch":
            return BestFirstSearch(**self.search_params)
        else:
            raise ValueError(f"Unknown search strategy {self.search_strategy}")

    def get_proof_search_heuristic(self):
        if self.proof_search_heuristic == "NegLoglikelihoodMinimizingHeuristic":
            return NegLoglikelihoodMinimizingHeuristic()
        else:
            raise ValueError(f"Unknown proof search heuristic {self.proof_search_heuristic}")

@dataclass_json
@dataclass
class EvalFile(object):
    path: str
    theorems: typing.Union[str, typing.List[str]]
    max_retry_attempts_limits: typing.Dict[str, int]
    max_time_limits_in_secs: typing.Dict[str, float]

@dataclass_json
@dataclass
class EvalDataset(object):
    project: str
    files: typing.List[EvalFile]
    negation: bool

@dataclass_json
@dataclass
class EvalBenchmark(object):
    name: str
    num_files: int
    language: ProofAction.Language
    datasets: typing.List[EvalDataset]
    few_shot_data_path_for_retrieval: str = None
    few_shot_metadata_filename_for_retrieval: str = None
    dfs_data_path_for_retrieval: str = None
    dfs_metadata_filename_for_retrieval: str = None
    timeout_per_theorem_in_secs: int = 720
    setup_cmds: typing.List[str] = field(default_factory=list)

@dataclass_json
@dataclass
class Experiments(object):
    env_settings: EnvSettings
    eval_settings: EvalSettings
    benchmark: EvalBenchmark

@dataclass_json
@dataclass
class EvalRunCheckpointInfo(object):
    checkpoint_file: str
    logging_dirs: typing.List[str]
    proof_dump_dir: str
    theorem_maps: typing.Dict[str, typing.Dict[str, bool]]

    def add_path_to_maps(self, path: str):
        if path not in self.theorem_maps:
            self.theorem_maps[path] = {}

    def add_theorem_to_maps(self, path: str, theorem: str, success: bool):
        self.theorem_maps[path][theorem] = success
        # First create a temporary copy of the checkpoint file
        # This is to avoid memory issues when writing to the file
        checkpoint_file = self.checkpoint_file + ".tmp"
        with open(checkpoint_file, "w") as f:
            f.write(self.to_json(indent=4))
        # Then move the temporary file to the original checkpoint file
        os.rename(checkpoint_file, self.checkpoint_file)
        # with open(self.checkpoint_file, "w") as f:
        #     f.write(self.to_json(indent=4))
    
@dataclass_json
@dataclass
class EvalProofResults(object):
    path: str
    theorem_map: typing.Dict[str, typing.Dict[str, ProofSearchResult]]

    def add_path_to_maps(self, path: str):
        if path not in self.theorem_map:
            self.theorem_map[path] = {}
    
    def add_theorem_to_maps(self, path: str, theorem: str, proof_result: ProofSearchResult):
        self.theorem_map[path][theorem] = proof_result
        # First create a temporary copy of the proof results file
        # This is to avoid memory issues when writing to the file
        proof_results_file = self.path + ".tmp"
        with open(proof_results_file, "w") as f:
            f.write(self.to_json(indent=4))
        # Then move the temporary file to the original proof results file
        os.rename(proof_results_file, self.path)
        # with open(self.path, "w") as f:
        #     f.write(self.to_json(indent=4))

@ray.remote
class EvalProofResultsActor(object):
    def __init__(self, eval_results: EvalProofResults):
        self.eval_results = eval_results
    
    def get_proof_results(self):
        return self.eval_results

    def get_theorem_map(self):
        return self.eval_results.theorem_map

    def add_path_to_maps(self, path: str):
        self.eval_results.add_path_to_maps(path)
    
    def add_theorem_to_maps(self, path: str, theorem: str, proof_result: ProofSearchResult):
        self.eval_results.add_theorem_to_maps(path, theorem, proof_result)

@ray.remote
class EvalRunCheckpointInfoActor(object):
    def __init__(self, eval_checkpoint: EvalRunCheckpointInfo):
        self.eval_checkpoint = eval_checkpoint
    
    def get_checkpoint_info(self):
        return self.eval_checkpoint

    def add_path_to_maps(self, path: str):
        self.eval_checkpoint.add_path_to_maps(path)
    
    def add_theorem_to_maps(self, path: str, theorem: str, success: bool):
        self.eval_checkpoint.add_theorem_to_maps(path, theorem, success)

def recursive_replace_keywords(cfg, key_word: str, replace_word: str):
    if isinstance(cfg, omegaconf.dictconfig.DictConfig) or isinstance(cfg, dict):
        keys = [key for key in cfg] # to avoid immutable dict error
        for key in keys:
            value = cfg[key]
            if isinstance(value, str):
                cfg[key] = value.replace(key_word, replace_word)
            elif isinstance(value, omegaconf.dictconfig.DictConfig) or \
                isinstance(value, omegaconf.listconfig.ListConfig) or \
                isinstance(value, dict) or \
                isinstance(value, list):
                recursive_replace_keywords(value, key_word, replace_word)
    elif isinstance(cfg, omegaconf.listconfig.ListConfig) or isinstance(cfg, list):
        for i in range(len(cfg)):
            value = cfg[i]
            if isinstance(value, str):
                cfg[i] = value.replace(key_word, replace_word)
            elif isinstance(value, omegaconf.dictconfig.DictConfig) or \
                isinstance(value, omegaconf.listconfig.ListConfig) or \
                isinstance(value, dict) or \
                isinstance(value, list):
                recursive_replace_keywords(value, key_word, replace_word)
    else:
        raise Exception(f"Invalid type: {type(cfg)}")


def parse_config(cfg):
    if "ROOT" in os.environ:
        root = os.environ["ROOT"]
    else:
        root = None
    if root is not None:
        # Replace all the <root> placeholders in all the paths in all the setting
        recursive_replace_keywords(cfg, "<root>", root)
    env_settings_cfg = cfg["env_settings"]
    env_settings = EnvSettings(
        name=env_settings_cfg["name"],
        retrieval_strategy=ProofEnvReRankStrategy(env_settings_cfg["retrieval_strategy"]))
    eval_settings_cfg = cfg["eval_settings"]
    eval_settings = EvalSettings(
        name=eval_settings_cfg["name"],
        use_hammer=eval_settings_cfg["use_hammer"],
        model_name=eval_settings_cfg["model_name"],
        is_seq2seq=eval_settings_cfg["is_seq2seq"],
        max_proof_depth=eval_settings_cfg["max_proof_depth"],
        max_seq_length=eval_settings_cfg["max_seq_length"],
        timeout_in_secs=eval_settings_cfg["timeout_in_secs"],
        proof_retries=eval_settings_cfg["proof_retries"],
        character_per_token=eval_settings_cfg["character_per_token"],
        width=eval_settings_cfg["width"],
        max_tokens_per_action=eval_settings_cfg["max_tokens_per_action"],
        temperature=eval_settings_cfg["temperature"],
        do_sample=eval_settings_cfg["do_sample"],
        top_k=eval_settings_cfg["top_k"],
        stop_tokens=eval_settings_cfg["stop_tokens"],
        padding=eval_settings_cfg["padding"],
        return_full_text=eval_settings_cfg["return_full_text"],
        compute_probabilities=eval_settings_cfg["compute_probabilities"],
        checkpoint_dir=eval_settings_cfg["checkpoint_dir"],
        should_checkpoint=eval_settings_cfg["should_checkpoint"],
        proof_dump_dir=eval_settings_cfg["proof_dump_dir"],
        use_human_readable_proof_context=eval_settings_cfg["use_human_readable_proof_context"],
        sample=eval_settings_cfg["sample"],
        sample_seed=eval_settings_cfg["sample_seed"],
        use_example_retrieval=eval_settings_cfg["use_example_retrieval"],
        always_use_useful_theorem_retrieval=eval_settings_cfg["always_use_useful_theorem_retrieval"],
        num_goal_per_prompt=eval_settings_cfg["num_goal_per_prompt"],
        search_strategy=eval_settings_cfg["search_strategy"],
        search_params=eval_settings_cfg["search_params"],
        proof_search_heuristic=eval_settings_cfg["proof_search_heuristic"],
        model_parallelism=eval_settings_cfg["model_parallelism"],
        do_lemmas_discovery=eval_settings_cfg["do_lemmas_discovery"] if "do_lemmas_discovery" in eval_settings_cfg else True,
        proof_tracer=ProofPathTracer.from_dict(eval_settings_cfg["proof_tracer"]) if "proof_tracer" in eval_settings_cfg else None
    )
    benchmark_cfg = cfg["benchmark"]
    datasets_cfg = benchmark_cfg["datasets"]
    eval_datasets = []
    for dataset_cfg in datasets_cfg:
        files_cfg = list(dataset_cfg["files"])
        eval_files = []
        for file_cfg in files_cfg:
            theorems = None
            if type(file_cfg["theorems"]) == str:
                theorems = file_cfg["theorems"]
            else:
                theorems = list(file_cfg["theorems"])
            if "max_retry_attempts_limits" not in file_cfg:
                max_retry_attempts_limits = {}
            else:
                max_retry_attempts_limits = file_cfg["max_retry_attempts_limits"]
            if "max_time_limits_in_secs" not in file_cfg:
                max_time_limits_in_secs = {}
            else:
                max_time_limits_in_secs = file_cfg["max_time_limits_in_secs"]
            eval_files.append(EvalFile(
                path=file_cfg["path"],
                theorems=theorems,
                max_retry_attempts_limits=max_retry_attempts_limits,
                max_time_limits_in_secs=max_time_limits_in_secs))
        eval_datasets.append(EvalDataset(
            project=dataset_cfg["project"],
            files=eval_files,
            negation=dataset_cfg["negation"] if "negation" in dataset_cfg else False))
    language = ProofAction.Language(benchmark_cfg["language"])
    benchmark = EvalBenchmark(
        name=benchmark_cfg["name"],
        num_files=benchmark_cfg["num_files"],
        language=language,
        datasets=eval_datasets,
        few_shot_data_path_for_retrieval=benchmark_cfg["few_shot_data_path_for_retrieval"],
        few_shot_metadata_filename_for_retrieval=benchmark_cfg["few_shot_metadata_filename_for_retrieval"],
        dfs_data_path_for_retrieval=benchmark_cfg["dfs_data_path_for_retrieval"],
        dfs_metadata_filename_for_retrieval=benchmark_cfg["dfs_metadata_filename_for_retrieval"],
        timeout_per_theorem_in_secs=benchmark_cfg["timeout_per_theorem_in_secs"] if "timeout_per_theorem_in_secs" in benchmark_cfg else 720)
    return Experiments(env_settings=env_settings, eval_settings=eval_settings, benchmark=benchmark)


#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
from proof_wala.llm_helpers.model import TrainingDataFormatterCallback
from proof_wala.llm_helpers.theorem_proving_training_dataset import TheoremProvingTrainingDataset
from itp_interface.tools.training_data import TrainingData, TrainingDataFormat
from proof_wala.parsers.proof_model_grammars import ProofModelGrammar, ProofModelPredGrammar
from proof_wala.parsers.grammars.prompt_template_grammar import PromptGrammar

class ProofModelTrainingDataset(TheoremProvingTrainingDataset):
    supported_training_data_configs = [
        'prompt__state__complete__steps_description_state',
        'prompt__state__complete__steps',
        'prompt__state__complete__steps_description',
        'prompt__state__complete__steps_description_state',
        'prompt__state_steps__complete__description',
        'prompt__state_steps__complete__state_description'
    ]
    ordering = {
        'prompt__state__complete__steps_description_state': 
            (['STATE'], ['PROOFSTEP', 'DESCRIPTION', 'STATE']),
        'prompt__state__complete__steps':
            (['STATE'], ['PROOFSTEP']),
        'prompt__state__complete__steps_description':
            (['STATE'], ['PROOFSTEP', 'DESCRIPTION']),
        'prompt__state__complete__steps_description_state':
            (['STATE'], ['PROOFSTEP', 'DESCRIPTION', 'STATE']),
        'prompt__state_steps__complete__state_description': 
            (['STATE', 'PROOFSTEP'], ['STATE', 'DESCRIPTION']),
        'prompt__state_steps__complete__description': 
            (['STATE', 'PROOFSTEP'], ['DESCRIPTION'])
    }
    def __init__(self, 
            training_data: TrainingData,
            characters_per_token: float = 3.6,
            max_tokens: int = 2048,
            state_delta: bool = False,
            max_distance_to_good: int = 10,
            training_data_gen_config: typing.List[str] = ['prompt__state__complete__steps_description_state']):
        assert len(training_data_gen_config) > 0, "Training data gen config is empty"
        assert characters_per_token >= 1, f"Characters per token must be at least 1, but is {characters_per_token}"
        super().__init__(training_data)
        self.prompt_grammar = PromptGrammar()
        self.characters_per_token = characters_per_token
        self.max_tokens = max_tokens
        self.max_tokens_in_prompt = self.max_tokens
        self.max_chars_in_prompt = int(self.max_tokens_in_prompt * self.characters_per_token)
        self.state_delta = state_delta
        self.training_data_gen_config = training_data_gen_config
        ordering_response, ordering_request = ProofModelTrainingDataset.ordering.get(training_data_gen_config[0], ProofModelTrainingDataset.supported_training_data_configs[0])
        self.response_grammar = ProofModelGrammar(state_delta=False, enable_proofstep=True, ordering=ordering_response) # Response grammar does not need state delta
        self.request_grammar = ProofModelPredGrammar(state_delta=self.state_delta, enable_proofstep=True, ordering=ordering_request)
        if self.max_tokens_in_prompt < 0:
            raise ValueError(f"Max tokens in prompt is negative: {self.max_tokens_in_prompt}, increase max tokens or decrease characters per token")
        self.prompt_delimiter = "\n"
        self.max_distance_to_good = max_distance_to_good
    
    def _reconstruct_proof_tree(self, 
        prev_state_id_map: typing.Dict[int, typing.Set[int]], 
        good_state_ids: typing.Set[int], 
        bad_state_ids: typing.Set[int],
        distance_map: typing.Dict[int, int]):
        distance = 1
        while True:
            # The idea is at some point no new good state ids will be added
            # This is level order traversal but in reverse
            new_good_state_ids = set()
            for end_state, start_states in prev_state_id_map.items():
                if end_state in good_state_ids:
                    for start_state in start_states:
                        if start_state not in good_state_ids:
                            # To avoid infinite loop we need to not add the start_state if it is already in good_state_ids
                            new_good_state_ids.add(start_state)
                            dist = distance_map.get(start_state, distance)
                            if dist >= distance:
                                distance_map[start_state] = distance
            distance += 1
            if len(new_good_state_ids) == 0:
                break
            good_state_ids.update(new_good_state_ids)
        # Now identify the states which are not good
        for end_state in prev_state_id_map.keys():
            if end_state not in good_state_ids:
                bad_state_ids.add(end_state)
    
    def _reconstruct_prev_state_id_map(self, training_datas: typing.List[TrainingDataFormat]) -> typing.Tuple[typing.Dict[int, int], int]:
            prev_state_id_map = {}
            done_state = None
            for training_data in training_datas:
                if training_data.addition_state_info is not None and len(training_data.addition_state_info) > 0:
                    start_state = training_data.addition_state_info.get("start_goal_id", None)
                    end_state = training_data.addition_state_info.get("end_goal_id", None)
                    if start_state is not None and end_state is not None:
                        prev_to_end_state = prev_state_id_map.get(end_state, set())
                        prev_to_end_state.add(start_state)
                        prev_state_id_map[end_state] = prev_to_end_state
                    if done_state is None and training_data.addition_state_info.get("done", False):
                        done_state = end_state
            return prev_state_id_map, done_state

    def load(self, **kwargs):
        super().load(**kwargs)
        # Go over each example and create a tree map for value function training
        proof_id_maps : typing.Dict[str, typing.List[TrainingDataFormat]] = {}
        self._cached_training_data = []
        for idx in range(len(self.training_data)):
            example = self.training_data[idx]
            training_datas : typing.List[TrainingDataFormat] = proof_id_maps.get(example.proof_id, [])
            training_datas.append(example)
            self._cached_training_data.append(example)
            proof_id_maps[example.proof_id] = training_datas
        for proof_id, training_datas in proof_id_maps.items():
            prev_state_id_map, done_state = self._reconstruct_prev_state_id_map(training_datas)
            # Now we have the prev_state_id_map and done_state
            # Every state from where we can reach done_state is a should have a good value
            # Every other state should have a bad value
            # First figure out all state ids from where we can reach done_state
            good_state_ids = set()
            good_state_ids.add(done_state)
            bad_state_ids = set()
            distance_map = {done_state: 0}
            self._reconstruct_proof_tree(prev_state_id_map, good_state_ids, bad_state_ids, distance_map)
            # Now we have the good_state_ids and bad_state_ids
            # Now annotate the training data with the value function
            for training_data in training_datas:
                if training_data.addition_state_info is not None and len(training_data.addition_state_info) > 0:
                    end_state_id = training_data.addition_state_info.get("end_goal_id", None)
                    if end_state_id is not None:
                        progress = training_data.addition_state_info.get("progress", "")
                        if end_state_id in good_state_ids and (progress == "StateChanged" or progress == "Done"):
                            distance = distance_map.get(end_state_id, self.max_distance_to_good)
                            distance = min(distance, self.max_distance_to_good)
                            progress = f"[GOOD] [{distance}] {progress}"
                        else:
                            progress = f"[BAD] {progress}"
                        training_data.addition_state_info["progress"] = progress

    def __len__(self):
        super_len = super().__len__()
        return super_len * len(self.training_data_gen_config)

    def __getitem__(self, idx):
        super_len = super().__len__()
        order_idx = idx // super_len
        idx = idx % super_len
        training_data_config = self.training_data_gen_config[order_idx]
        ordering_response, ordering_request = ProofModelTrainingDataset.ordering.get(
            training_data_config, 
            ProofModelTrainingDataset.supported_training_data_configs[0])
        self.response_grammar.ordering = ordering_response
        self.request_grammar.ordering = ordering_request
        if hasattr(self, "_cached_training_data"):
            proof_state = self._cached_training_data[idx]
        else:
            proof_state = self.training_data[idx]
        prompt = self.response_grammar.format_as_per_grammar(proof_state, max_token_cnt=self.max_tokens_in_prompt, characters_per_token=self.characters_per_token)
        completion = self.request_grammar.format_as_per_grammar(proof_state, max_token_cnt=self.max_tokens_in_prompt, characters_per_token=self.characters_per_token)
        prompt += self.prompt_delimiter
        return {
            "prompt": prompt,
            "completion": completion
        }
    
    # def prompt_formatter(
    #         response: CoqGptResponse, 
    #         max_tokens_in_prompt: int = None, 
    #         characters_per_token: float = 4):
    #     prompt = ProofModelTrainingDataset.response_grammar.format_as_per_grammar(
    #         response, 
    #         max_token_cnt=max_tokens_in_prompt, 
    #         characters_per_token=characters_per_token)
    #     return prompt

    # def response_parser(request_str: str) -> CoqGptRequest:
    #     if request_str.endswith(ProofModelPromptTrainingDataFormatter.stopping_token):
    #         request_str = request_str[:-len(ProofModelPromptTrainingDataFormatter.stopping_token)]
    #     request, _ = ProofModelTrainingDataset.request_grammar.attempt_parsing(request_str)
    #     return request
    

class ProofModelPromptTrainingDataFormatter(TrainingDataFormatterCallback):
    stopping_token = "[END]"
    def __call__(self, training_data_format_examples: typing.List[typing.Dict[str, str]]) -> typing.List[str]:
        output_texts = []
        all_prompts = training_data_format_examples["prompt"]
        all_completions = training_data_format_examples["completion"]
        stopping_token = ""
        for prompt, completion in zip(all_prompts, all_completions):
            output_text = f"{prompt}{completion}{stopping_token}"
            output_texts.append(output_text)
        return output_texts
    
    def get_prompt_and_completion(self, training_data_examples: typing.List[typing.Dict[str, str]]) -> typing.List[typing.Tuple[str, str]]:
        all_prompts = training_data_examples["prompt"]
        all_completions = training_data_examples["completion"]
        prompt_and_completions = []
        if isinstance(all_prompts, str):
            all_prompts = [all_prompts]
            assert isinstance(all_completions, str)
            all_completions = [all_completions]
        assert len(all_prompts) == len(all_completions)
        for prompt, completion in zip(all_prompts, all_completions):
            # Trim everything after the stopping token from the completion
            completion = completion.split(self.get_stopping_token())[0]
            prompt_and_completion = (prompt, completion + self.get_stopping_token())
            prompt_and_completions.append(prompt_and_completion)
        return prompt_and_completions
    
    def get_stopping_token(self):
        return ProofModelPromptTrainingDataFormatter.stopping_token

if __name__ == "__main__":
    data_folder = f".log/traces"
    meta_filename = "local.meta.json"
    training_data = TrainingData(data_folder, meta_filename)
    # Not removing system_prompt_file and conversation_prompt_file though not needed
    # as removing makes the function signature incompatable with the call-type in run_training.py
    training_data_gen_config = ProofModelTrainingDataset.supported_training_data_configs
    with ProofModelTrainingDataset(training_data, max_tokens=4096, state_delta=True, training_data_gen_config=training_data_gen_config) as dataset:
        hf_dataset = dataset.get_hf_dataset()
        formatter = ProofModelPromptTrainingDataFormatter()
        formatted_dataset = formatter(hf_dataset)
        for example_idx, example in enumerate(hf_dataset):
            prompt = example["prompt"]
            completion = example["completion"]
            print(f"[{example_idx + 1}] Prompt:\n{prompt}")
            print('-'*20)
            print(f"Completion:\n{completion}")
            print('-'*20)
            print('='*20)
        prompt_and_completions = formatter.get_prompt_and_completion(hf_dataset)
    pass


import ray
from torch.utils.data import Dataset
from datasets import Dataset as HFDataset
from itp_interface.tools.training_data import TrainingData

class TheoremProvingTrainingDataset(Dataset):
    def __init__(self, training_data: TrainingData):
        self.training_data = training_data
    
    def load(self, **kwargs):
        self.training_data.load()
        ray.shutdown() # Force shutdown ray because everything is in memory now.
        # execute ray stop in terminal to stop ray

    def unload(self):
        self.training_data.unload()

    def __enter__(self):
        self.load()
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.unload()
        pass

    def __len__(self):
        assert self.training_data._is_loaded, "Training data not loaded"
        return len(self.training_data)
    
    def __getitem__(self, idx):
        assert self.training_data._is_loaded, "Training data not loaded"
        example = self.training_data[idx]
        goals = []
        for idx, goal in enumerate(example.start_goals):
            if len(goal.hypotheses) > 0:
                hyps = '\n'.join(goal.hypotheses)
                goals.append(f"[GOAL] {idx + 1}\n{goal.goal}\n[HYPOTHESES] {idx + 1}\n{hyps}")
            else:
                goals.append(f"[GOAL] {idx + 1}\n{goal.goal}")
        goals = '\n'.join(goals)
        goals = goals + '\n[PROOFSTEP]\n'
        proofstep = '\n'.join(example.proof_steps)
        proofstep = f"{proofstep}"
        return {
            "goals": goals,
            "proofstep": proofstep
        }
    
    def get_hf_dataset(self):
        assert self.training_data._is_loaded, "Training data not loaded"
        return HFDataset.from_list(self)


import heapq
import os
import typing
import time
import logging

from multiprocessing import Pool
try:
    from .search import Node, Edge, SearchAlgorithm
except ImportError:
    from search import Node, Edge, SearchAlgorithm


class BestFirstSearch(SearchAlgorithm):
    def __init__(self):
        super().__init__()

    def search(
            self,
            start: Node, 
            goal: Node,
            heuristic: typing.Callable[[Node, Edge, Node], float], 
            generate_children: typing.Callable[[Node, float], typing.Tuple[typing.List[Node], typing.List[Edge]]] = None,
            parallel_count: int = None,
            build_tree: bool = True,
            timeout_in_secs: float = None,
            logger : logging.Logger = None) -> typing.Tuple[Node, bool, float]:
        logger = logger if logger else logging.getLogger(__name__)
        assert (generate_children is not None and build_tree) or not build_tree, "Must provide generate_children function"
        time_elapsed = 0
        start_time = time.time()
        timeout_in_secs = timeout_in_secs if timeout_in_secs else float('inf')
        start.cummulative_score = 0
        start.distance_from_root = 0
        frontier : typing.List[typing.Tuple[float, Node]] = [(heuristic(None, None, start), start)]
        explored = {start: start}  # Now keeping track of explored nodes directly
        heapq.heapify(frontier) # this is a min heap

        # Initialize multiprocessing pool
        pool = Pool(processes=parallel_count if parallel_count else max(os.cpu_count() - 1, 1))
        tree_nodes = set([start])  # Keep track of all nodes in the tree
        while frontier:
            _, current_node = heapq.heappop(frontier) # Pop the node with the lowest score
            # print(f"Current distance from root: {current_node.distance_from_root}, cummulative score: {current_node.cummulative_score}")
            if current_node == goal:
                pool.close()
                pool.join()
                assert len(current_node.parents) > 0, f"Goal node must have at least one parent"
                return (current_node, True, time_elapsed)

            if time_elapsed > timeout_in_secs:
                pool.close()
                pool.join()
                return (start, False, time_elapsed)

            if build_tree:
                remaining_timeout = timeout_in_secs - time_elapsed
                children, edges = generate_children(current_node, remaining_timeout)
            else:
                children = [(child, edge) for child, edge in zip(current_node.children, current_node.edges)]
                children = current_node.children
                edges = current_node.edges

            # All costs should be computed before adding to the tree
            child_costs = pool.starmap(heuristic, [[current_node, edge, node] for node, edge in zip(children, edges)])                                

            explorable_children_with_costs = {}
            for child, edge, cost in zip(children, edges, child_costs):
                if child not in explored:
                    if build_tree:
                        child.distance_from_root = min(child.distance_from_root, current_node.distance_from_root + 1)
                        child.cummulative_score = min(child.cummulative_score, cost)
                        current_node.add_child(child, edge)  # Add child to current node's children list
                    explored[child] = child  # Add child to explored set
                    if child not in explorable_children_with_costs:
                        explorable_children_with_costs[child] = cost
                    else:
                        explorable_children_with_costs[child] = min(explorable_children_with_costs[child], cost)
                elif build_tree:
                    child_idx = None
                    try:
                        child_idx = current_node.children.index(child)
                    except ValueError:
                        pass
                    if child_idx is None:
                        explored_child = explored[child]
                        explored_child.distance_from_root = min(explored_child.distance_from_root, current_node.distance_from_root + 1)
                        explored_child.cummulative_score = min(explored_child.cummulative_score, cost)
                        current_node.add_child(explored_child, edge)
                    elif edge != current_node.edges[child_idx]:
                        if edge not in current_node.edges[child_idx].equivalent_edges:
                            current_node.edges[child_idx].add_equivalent_edge(edge)
                            explored_child = current_node.children[child_idx]
                            explored_child.distance_from_root = min(explored_child.distance_from_root, current_node.distance_from_root + 1)
                            if cost < explored_child.cummulative_score:
                                if explored_child not in explorable_children_with_costs:
                                    explorable_children_with_costs[explored_child] = cost
                                else:
                                    explorable_children_with_costs[explored_child] = min(explorable_children_with_costs[explored_child], cost)
                            explored_child.cummulative_score = min(explored_child.cummulative_score, cost)

            for child, cost in explorable_children_with_costs.items():
                if child not in tree_nodes:
                    heapq.heappush(frontier, (cost, child))
                    tree_nodes.add(child) # Don't re-add the same node to the frontier
            
            time_elapsed = time.time() - start_time

        pool.close()
        pool.join()
        return (start, False, time_elapsed)


#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
from proof_wala.llm_helpers.model import TrainingDataFormatterCallback

class BasicTrainingDataFormatterCallback(TrainingDataFormatterCallback):
    def __init__(self):
        pass
    
    def __call__(self, training_data_format_examples: typing.List[typing.Dict[str, str]]) -> typing.List[str]:
        output_texts = []
        all_goals = training_data_format_examples["goals"]
        all_proofsteps = training_data_format_examples["proofstep"]
        stopping_token = self.get_stopping_token()
        for goals, proofstep in zip(all_goals, all_proofsteps):
            output_text = f"{goals}{proofstep}{stopping_token}"
            output_texts.append(output_text)
        return output_texts
    
    def get_prompt_and_completion(self, training_data_examples: typing.List[typing.Dict[str, str]]) -> typing.List[typing.Tuple[str, str]]:
        all_goals = training_data_examples["goals"]
        all_proofsteps = training_data_examples["proofstep"]
        prompt_and_completions = []
        if isinstance(all_goals, str):
            all_goals = [all_goals]
            assert isinstance(all_proofsteps, str)
            all_proofsteps = [all_proofsteps]
        assert len(all_goals) == len(all_proofsteps)
        for goals, proofstep in zip(all_goals, all_proofsteps):
            prompt_and_completion = (goals, proofstep + self.get_stopping_token())
            prompt_and_completions.append(prompt_and_completion)
        return prompt_and_completions
    
    def get_stopping_token(self):
        return "[END]"


import subprocess
import unittest
import os

class TestProofWalaSearchCommand(unittest.TestCase):
    def test_proof_wala_search_command(self):
        """
        Test that the 'proof-wala-search' command runs successfully with the given configuration.
        """
        # Construct the command as a single string.
        command = (
            "proof-wala-search --config-dir=src/proof_wala/main/config "
            "--config-name=eval_simple_lean_test_multilingual_easy.yaml"
        )

        try:
            # Run the command using shell=True so that the shell does the PATH lookup.
            result = subprocess.run(
                command,
                shell=True,
                capture_output=True,
                text=True,
                timeout=700
            )
        except subprocess.TimeoutExpired as e:
            self.fail(f"'proof-wala-search' command timed out: {e}")
        except Exception as e:
            self.fail(f"Error running 'proof-wala-search': {e}")

        # Check that the command exited with a return code of 0.
        self.assertEqual(
            result.returncode, 0,
            msg=f"'proof-wala-search' failed with return code {result.returncode}. Stderr: {result.stderr}"
        )

        # Print all the files in the .log/proofs_logs/simple_lean_test_easy/simple_lean_test
        # directory to see what was generated.
        # Do a list and pick the last folder in the list as per the sorted order
        dirs = sorted(os.listdir(".log/proofs_logs/simple_lean_test_easy/simple_lean_test"))
        print(dirs)
        last_dir = dirs[-1]
        gpu_dir = os.path.join(".log/proofs_logs/simple_lean_test_easy/simple_lean_test", last_dir, "gpu_None_0")
        log_file = os.path.join(gpu_dir, "eval_dataset_multiple_attempts.log")
        print("Log file:", log_file)
        with open(log_file, "r") as f:
            print(f.read())

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
import logging
import numpy as np
import os
from datetime import datetime
from proof_wala.proof_search.search_driver import ProofActionGenerator, ProofSearhHeuristic, ProofSearchDriver, ProofStateInfo, ProofPathTracer
from itp_interface.rl.simple_proof_env import ProofState, ProofAction, ProofEnv
from itp_interface.rl.simpl_proof_env_pool import replicate_proof_env
from itp_interface.tools.proof_exec_callback import ProofExecutorCallback
from itp_interface.tools.training_data_format import TrainingDataMetadataFormat
from itp_interface.tools.training_data import TrainingData
from proof_wala.search.search import Node, SearchAlgorithm, Edge
from proof_wala.search.best_first_search import BestFirstSearch
from proof_wala.search.beam_search import BeamSearch
os.environ["CUDA_VISIBLE_DEVICES"] = "5"

class RandomCoqProofActionGenerator(ProofActionGenerator):
    def __init__(self, width: int = 4):
        super().__init__(width)
        pass

    def get_proof_end_for_language(self, language: ProofAction.Language) -> ProofAction:
        return ProofAction(ProofAction.ActionType.RUN_TACTIC, ProofAction.Language.COQ, tactics=['Qed.'])
    
    def generate_actions(self, state_info: ProofStateInfo, k: int = None) -> typing.List[typing.Tuple[float, ProofAction]]:
        state : ProofState = state_info.proof_state
        done: bool = state_info.done
        assert state.language == ProofAction.Language.COQ, "Only Coq is supported"
        if done:
            return []
        elif len(state.proof_tree) == 0:
            tactics = {
                "intros.": 0.3,
                "intro.": 0.5,
                "simpl.": 0.20
            }
        elif len(state.training_data_format.start_goals) == 0:
            tactics = {
                "Qed.": 1.0
            }
        else:
            tactics = {
                "Qed.": 0.2,
                "intro.": 0.15,
                "intros.": 0.10,
                "trivial.": 0.15,
                "reflexivity.": 0.10,
                "auto.": 0.15,
                "firstorder.": 0.15,
            }
        # Generate random actions
        actions = []
        # Sample actions based on the probabilities
        k = k if k else self.width
        for _ in range(k):
            action = np.random.choice(list(tactics.keys()), p=list(tactics.values()))
            actions.append((tactics[action], ProofAction(ProofAction.ActionType.RUN_TACTIC, state.language, tactics=[action])))
        return actions

class ProofFoundHeuristic(ProofSearhHeuristic):
    def __init__(self):
        pass

    def __call__(self, parent_node: Node, edge: Edge, node: Node) -> float:
        state_info : ProofStateInfo = node.other_data
        state : ProofState = state_info.proof_state
        if state_info.done or len(state.training_data_format.start_goals) == 0:
            return -100.0
        else:
            return node.score

def test_search_algorithm(
        exp_name: str,
        algo: SearchAlgorithm, 
        proof_env: ProofEnv,
        proof_search_heuristic: ProofFoundHeuristic,
        action_generator: ProofActionGenerator,
        search_width: int = 4, 
        attempt_count: int = 1,
        timeout_in_secs: float = 60,
        trace_settings: ProofPathTracer = None,
        extract_original_proof: bool = False) -> typing.Callable[[Node], typing.List[Node]]:
    #tracemalloc.start()
    proof_found = False
    tree_dump_folder = f".log/proof_search/{exp_name}"
    time_now = datetime.now().strftime("%Y%m%d_%H%M%S")
    tree_dump_folder = os.path.join(tree_dump_folder, time_now)
    os.makedirs(tree_dump_folder, exist_ok=True)
    max_attempts = attempt_count
    final_proof_res = None
    while attempt_count > 0 and not proof_found:
        print(f"Attempt {max_attempts - attempt_count + 1} of {max_attempts}")
        search_driver = ProofSearchDriver(
            algo,
            action_generator,
            proof_search_heuristic,
            width=search_width,
            tracer=trace_settings)
        replicated_env = replicate_proof_env(proof_env)
        start_node, tree_node, proof_res = search_driver.search_proof(
            proof_env,
            timeout_in_secs=timeout_in_secs,
            extract_original_proofs=extract_original_proof)
        final_proof_res = proof_res
        proof_paths = []
        if proof_res.proof_found:
            proof_paths = algo.reconstruct_all_paths(start_node, tree_node)
            print(f"{len(proof_paths)} proof(s) found in {max_attempts - attempt_count + 1} attempt(s)")
            if len(proof_paths) == 0:
                print("Proof path is empty, something went wrong!")
                print(proof_res)
            else:
                # Run the proof steps and check if the proof is finished
                with replicated_env:
                    for td in proof_res.proof_steps:
                        replicated_env.step(ProofAction(ProofAction.ActionType.RUN_TACTIC, proof_env.language, tactics=td.proof_steps))
                    assert replicated_env.done
                proof_found = True
        dot = algo.visualize_search(start_node, show=False, mark_paths=proof_paths)
        dot.render(os.path.join(tree_dump_folder, f"proof_search_{max_attempts - attempt_count + 1}"), format='png', quiet=True)
        attempt_count -= 1
    print(final_proof_res)

if __name__ == '__main__':
    import ray
    ray.init(num_cpus=10, num_gpus=1, ignore_reinit_error=True)
    proof_exec_callback = ProofExecutorCallback(
        project_folder=".",
        file_path="src/proof_wala/data/proofs/coq/simple2/thms.v"
    )
    theorem_names = [
        "nat_add_comm",
        "double_neg",
        "trival_implication",
        "modus_ponens",
        "modus_tollens",
        "disjunctive_syllogism",
        "contrapositive",
        "nat_zero_add",
        "nat_add_zero",
        "nat_add_succ",
        "nat_succ_add"
    ]
    width = 10
    tracer = ProofPathTracer(
        collect_traces=True,
        folder=".log/traces",
        training_meta_filename="traces_meta.json",
        training_metadata=TrainingDataMetadataFormat(
            data_filename_prefix="traces_data",
            lemma_ref_filename_prefix="traces_lemma_ref"
        ),
        max_parallelism=4
    )
    os.makedirs(tracer.folder, exist_ok=True)
    tracer.reset_training_data()
    for search_aglo in [BestFirstSearch(), BeamSearch(3)]:
        algo_name = search_aglo.__class__.__name__
        print(f"Running tests for {algo_name}")
        for theorem_name in theorem_names:
            print(f"Trying to prove {theorem_name}")
            language = ProofAction.Language.COQ
            always_retrieve_thms = False
            logger = logging.getLogger(__name__)
            env = ProofEnv("test", proof_exec_callback, theorem_name, max_proof_depth=10, always_retrieve_thms=always_retrieve_thms, logger=logger)
            test_search_algorithm(
                exp_name=os.path.join(algo_name, theorem_name),
                algo=search_aglo,
                proof_env=env,
                proof_search_heuristic=ProofFoundHeuristic(),
                action_generator=RandomCoqProofActionGenerator(width),
                search_width=width,
                attempt_count=2,
                timeout_in_secs=60,
                trace_settings=tracer,
                extract_original_proof=True)
            print('-' * 80)
        print('=' * 80)
    tracer.save()


#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import json
import os
import time
import uuid
from itp_interface.tools.training_data import TrainingData
from itp_interface.tools.log_utils import setup_logger
from proof_wala.llm_helpers.model import Model
from proof_wala.main.config import Experiment, ExperimentType

def train_experiment(experiment: Experiment, time_now: str = None):
    assert experiment.expertiment_type == ExperimentType.Training, f"Experiment type must be {ExperimentType.Training}, but is {experiment.expertiment_type}"
    # assert experiment.training_data_settings.training_data_dir is not None, f"Training data directory must be specified"
    time_now = time.strftime("%Y%m%d-%H%M%S") if time_now is None else time_now
    training_dataset_log_dir = os.path.join(experiment.training_data_settings.training_data_log_dir, time_now)
    os.makedirs(training_dataset_log_dir, exist_ok=True)
    eval_dataset_log_dir = os.path.join(experiment.training_data_settings.eval_data_log_dir, time_now) if experiment.training_data_settings.eval_data_log_dir is not None else None
    if eval_dataset_log_dir is not None:
        os.makedirs(eval_dataset_log_dir, exist_ok=True)
    test_dataset_log_dir = os.path.join(experiment.training_data_settings.test_data_log_dir, time_now) if experiment.training_data_settings.test_data_log_dir is not None else None
    if test_dataset_log_dir is not None:
        os.makedirs(test_dataset_log_dir, exist_ok=True)
    should_train = experiment.training_data_settings.training_data_dir is not None
    should_load_eval = experiment.training_data_settings.eval_data_dir is not None or experiment.training_data_settings.evals is not None
    should_load_test = experiment.training_data_settings.test_data_dir is not None
    should_split_train_eval = experiment.training_settings.train_eval_split and should_train
    if should_split_train_eval:
        assert not should_load_eval, f"Cannot split train and eval data if eval data is loaded from {experiment.training_data_settings.eval_data_dir}"
    # Load the training data
    gpuid_local = os.environ.get("LOCAL_RANK", None)
    gpuid_global = os.environ.get("RANK", None)
    gpuid_unique = str(uuid.uuid4())
    gpuid = f"{gpuid_local}-{gpuid_global}-{gpuid_unique}"
    hf_training_dataset = None
    training_dataset_type = experiment.training_data_settings.training_dataset_type.get_class()
    if should_train:
        training_data = TrainingData(experiment.training_data_settings.training_data_dir, experiment.training_data_settings.training_meta_filename, logger=setup_logger(f"TrainingData-{gpuid}", os.path.join(training_dataset_log_dir, f"training_data-{gpuid}.log")))
        with training_dataset_type(training_data, **experiment.training_data_settings.training_dataset_args) as training_dataset:
            hf_training_dataset = training_dataset.get_hf_dataset()
            if should_split_train_eval:
                hf_train_test_split = hf_training_dataset.train_test_split(test_size=experiment.training_settings.eval_percentage, seed=experiment.training_settings.training_args.data_seed)
                hf_training_dataset = hf_train_test_split["train"]
                hf_eval_dataset = hf_train_test_split["test"]
            else:
                hf_eval_dataset = None
            if experiment.training_settings.train_percentage < 1.0 and not should_split_train_eval:
                # Sample train_percentage of the dataset
                hf_training_dataset = hf_training_dataset.shuffle(seed=experiment.training_settings.training_args.data_seed).select(range(int(len(hf_training_dataset) * experiment.training_settings.train_percentage)))
    # Load the eval data
    if should_load_eval:
        if experiment.training_data_settings.evals is None:
            eval_data = TrainingData(experiment.training_data_settings.eval_data_dir, experiment.training_data_settings.eval_meta_filename, logger=setup_logger("EvalData-{gpuid}", os.path.join(eval_dataset_log_dir, f"eval_data-{gpuid}.log")))
            with training_dataset_type(eval_data, **experiment.training_data_settings.training_dataset_args) as eval_dataset:
                hf_eval_dataset = eval_dataset.get_hf_dataset()
        else:
            for eval_settings in experiment.training_data_settings.evals:
                os.makedirs(eval_settings.eval_data_log_dir, exist_ok=True) # Create the log directories
            eval_datasets = {
                eval_settings.eval_name: TrainingData(eval_settings.eval_data_dir, eval_settings.eval_meta_filename, logger=setup_logger(f"EvalData-{eval_settings.eval_name}-{gpuid}", os.path.join(eval_settings.eval_data_log_dir, f"eval_data-{eval_settings.eval_name}-{gpuid}.log")))
                for eval_settings in experiment.training_data_settings.evals
            }
            hf_eval_datasets = {}
            for eval_name, eval_data in eval_datasets.items():
                with training_dataset_type(eval_data, **experiment.training_data_settings.training_dataset_args) as eval_dataset:
                    hf_eval_datasets[eval_name] = eval_dataset.get_hf_dataset()
            hf_eval_dataset = hf_eval_datasets
    else:
        hf_eval_dataset = hf_eval_dataset if should_split_train_eval else None
    if hf_eval_dataset is not None:
        if experiment.training_settings.eval_percentage < 1.0 and not should_split_train_eval:
            # Sample eval_percentage of the dataset
            if isinstance(hf_eval_dataset, dict):
                for eval_name, eval_dataset in hf_eval_dataset.items():
                    hf_eval_dataset[eval_name] = eval_dataset.shuffle(seed=experiment.training_settings.training_args.data_seed).select(range(int(len(eval_dataset) * experiment.training_settings.eval_percentage)))
            else:
                hf_eval_dataset = hf_eval_dataset.shuffle(seed=experiment.training_settings.training_args.data_seed).select(range(int(len(hf_eval_dataset) * experiment.training_settings.eval_percentage)))
    # Load the test data
    if should_load_test:
        test_data = TrainingData(experiment.training_data_settings.test_data_dir, experiment.training_data_settings.test_meta_filename, logger=setup_logger("TestData-{gpuid}", os.path.join(test_dataset_log_dir, f"test_data-{gpuid}.log")))
        with training_dataset_type(test_data, **experiment.training_data_settings.training_dataset_args) as test_dataset:
            hf_test_dataset = test_dataset.get_hf_dataset()
    else:
        hf_test_dataset = None
    if hf_test_dataset is not None:
        if experiment.training_settings.test_percentage < 1.0:
            # Sample test_percentage of the dataset
            hf_test_dataset = hf_test_dataset.shuffle(seed=experiment.training_settings.training_args.data_seed).select(range(int(len(hf_test_dataset) * experiment.training_settings.test_percentage)))
    model_log_dir = os.path.join(experiment.model_settings.logging_dir, time_now)
    os.makedirs(model_log_dir, exist_ok=True)
    model_args = experiment.model_settings.model_args
    if "token" not in model_args or model_args["token"] is None:
        # try to get from .secrets/huggingface_token.json
        if os.path.exists(".secrets/huggingface_token.json"):
            with open(".secrets/huggingface_token.json", "r") as f:
                model_args["token"] = json.load(f)["token"]
    if "comet_experiment" not in model_args or model_args["comet_experiment"] is None:
        model_args["comet_experiment"] = f"{time_now}-{experiment.name}"
    # Train the model
    if experiment.training_settings.training_args.output_dir is None or experiment.training_settings.training_args.output_dir == "":
        experiment.training_settings.training_args.output_dir = os.path.join(model_log_dir, experiment.name)
    else:
        experiment.training_settings.training_args.output_dir = os.path.join(experiment.training_settings.training_args.output_dir, experiment.name)
    os.makedirs(experiment.training_settings.training_args.output_dir, exist_ok=True)
    model = Model(
        experiment.model_settings.name_or_path, 
        experiment.training_settings.training_args, 
        model_log_dir,
        **experiment.model_settings.model_args)
    training_data_formatter_callback = experiment.training_data_settings.training_data_formatter_type.get_class()
    with model:
        if hf_training_dataset is None:
            hf_training_dataset = hf_test_dataset
        if hf_eval_dataset is not None:
            eval_set = hf_eval_dataset
        else:
            eval_set = None
        model.train(
            time_now,
            training_data_formatter_callback(),
            hf_training_dataset, 
            eval_set)
        if hf_test_dataset is not None:
            model.training_args.do_train = False
            model.training_args.do_eval = True
            model.train(
                time_now,
                training_data_formatter_callback(),
                hf_training_dataset, 
                hf_test_dataset)


#!/usr/bin/env python3

import sys

root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import os
import typing
import omegaconf
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from enum import Enum
from transformers import TrainingArguments
from proof_wala.llm_helpers.theorem_proving_training_dataset import TheoremProvingTrainingDataset
from proof_wala.itp.training_data_formatter import BasicTrainingDataFormatterCallback
from proof_wala.itp.copra_training_data_formatter import CopraPromptTrainingDataFormatter, CopraTrainingDataset
from proof_wala.itp.codet5_training_data_formatter import CodeT5PromptTrainingDataFormatter, CodeT5TrainingDataset
from proof_wala.itp.proof_model_training_data_formatter import ProofModelTrainingDataset, ProofModelPromptTrainingDataFormatter

class ExperimentType(Enum):
    Training = "Training"
    Inferencing = "Inferencing"
    Evaluation = "Evaluation"
    TokenCount = "TokenCount"

    def __str__(self):
        return self.value

class TrainingDatasetType(Enum):
    TheoremProvingTrainingDataset = "TheoremProvingTrainingDataset"
    CopraTrainingDataset = "CopraTrainingDataset"
    CodeT5TrainingDataset = "CodeT5TrainingDataset"
    ProofModelTrainingDataset = "ProofModelTrainingDataset" 

    def __str__(self):
        return self.value
    
    def get_class(self):
        if self == TrainingDatasetType.TheoremProvingTrainingDataset:
            return TheoremProvingTrainingDataset
        elif self == TrainingDatasetType.CopraTrainingDataset:
            return CopraTrainingDataset
        elif self == TrainingDatasetType.CodeT5TrainingDataset:
            return CodeT5TrainingDataset
        elif self == TrainingDatasetType.ProofModelTrainingDataset:
            return ProofModelTrainingDataset
        else:
            raise Exception(f"Invalid training dataset type: {self}")
    
class TrainingDataFormatterType(Enum):
    BasicTrainingDataFormatterCallback = "BasicTrainingDataFormatterCallback"
    CopraPromptTrainingDataFormatter = "CopraPromptTrainingDataFormatter"
    CodeT5PromptTrainingDataFormatter = "CodeT5PromptTrainingDataFormatter"
    ProofModelPromptTrainingDataFormatter = "ProofModelPromptTrainingDataFormatter"

    def __str__(self):
        return self.value
    
    def get_class(self):
        if self == TrainingDataFormatterType.BasicTrainingDataFormatterCallback:
            return BasicTrainingDataFormatterCallback
        elif self == TrainingDataFormatterType.CopraPromptTrainingDataFormatter:
            return CopraPromptTrainingDataFormatter
        elif self == TrainingDataFormatterType.CodeT5PromptTrainingDataFormatter:
            return CodeT5PromptTrainingDataFormatter
        elif self == TrainingDataFormatterType.ProofModelPromptTrainingDataFormatter:
            return ProofModelPromptTrainingDataFormatter
        else:
            raise Exception(f"Invalid training data formatter type: {self}")

@dataclass_json
@dataclass
class ModelSettings(object):
    name_or_path: str
    logging_dir: str
    model_args: dict = field(default_factory=dict)

@dataclass_json
@dataclass
class EvalSettings(object):
    eval_name: str
    eval_data_dir: str
    eval_meta_filename: str
    eval_data_log_dir: str

@dataclass_json
@dataclass
class TrainingDataSettings(object):
    training_data_dir: str
    training_meta_filename: str
    training_data_log_dir: str
    training_dataset_type: TrainingDatasetType
    training_data_formatter_type: TrainingDataFormatterType
    training_dataset_args: dict = field(default_factory=dict)   
    eval_data_dir: typing.Optional[str] = None
    eval_meta_filename: typing.Optional[str] = None
    eval_data_log_dir: typing.Optional[str] = None
    test_data_dir: typing.Optional[str] = None
    test_meta_filename: typing.Optional[str] = None
    test_data_log_dir: typing.Optional[str] = None
    evals : typing.Optional[typing.List[EvalSettings]] = None

@dataclass_json
@dataclass
class TrainingSettings(object):
    train_eval_split: bool = False
    train_percentage: float = 1.0
    eval_percentage: float = 1.0
    test_percentage: float = 1.0
    training_args: TrainingArguments = field(default_factory=TrainingArguments)

@dataclass_json
@dataclass
class Experiment(object):
    name : str
    expertiment_type: ExperimentType
    model_settings: ModelSettings
    training_data_settings: TrainingDataSettings
    training_settings: TrainingSettings

def recursive_replace_keywords(cfg, key_word: str, replace_word: str):
    if isinstance(cfg, omegaconf.dictconfig.DictConfig) or isinstance(cfg, dict):
        keys = [key for key in cfg] # to avoid immutable dict error
        for key in keys:
            value = cfg[key]
            if isinstance(value, str):
                cfg[key] = value.replace(key_word, replace_word)
            elif isinstance(value, omegaconf.dictconfig.DictConfig) or \
                isinstance(value, omegaconf.listconfig.ListConfig) or \
                isinstance(value, dict) or \
                isinstance(value, list):
                recursive_replace_keywords(value, key_word, replace_word)
    elif isinstance(cfg, omegaconf.listconfig.ListConfig) or isinstance(cfg, list):
        for i in range(len(cfg)):
            value = cfg[i]
            if isinstance(value, str):
                cfg[i] = value.replace(key_word, replace_word)
            elif isinstance(value, omegaconf.dictconfig.DictConfig) or \
                isinstance(value, omegaconf.listconfig.ListConfig) or \
                isinstance(value, dict) or \
                isinstance(value, list):
                recursive_replace_keywords(value, key_word, replace_word)
    else:
        raise Exception(f"Invalid type: {type(cfg)}")


def parse_config(cfg) -> Experiment:
    if "ROOT" in os.environ:
        root = os.environ["ROOT"]
    else:
        root = None
    if root is not None:
        # Replace all the <root> placeholders in all the paths in all the setting
        recursive_replace_keywords(cfg, "<root>", root)
    name = cfg["name"]
    experiment_type = ExperimentType(cfg["experiment_type"])
    model_settings : ModelSettings = ModelSettings.from_dict(cfg["model_settings"])
    if model_settings.model_args is None:
        model_settings.model_args = {}
    training_data_settings : TrainingDataSettings = TrainingDataSettings.from_dict(cfg["training_data_settings"])
    if training_data_settings.training_dataset_args is None:
        training_data_settings.training_dataset_args = {}
    training_settings = TrainingSettings.from_dict(cfg["training_settings"])
        
    experiment = Experiment(
        name=name,
        expertiment_type=experiment_type,
        model_settings=model_settings,
        training_data_settings=training_data_settings,
        training_settings=training_settings
    )
    return experiment


#!/usr/bin/env python3
import typing

class Llama2ChatFormat(object):
    """
        prompt  = f"<<SYS>>\\n{system}\\n<</SYS>>\\n\\n{user_1}"
        prompt += f"<s>[INST] {prompt.strip()} [/INST] {answer_1.strip()} </s>"
        prompt += f"<s>[INST] {user_2.strip()} [/INST] {answer_2.strip()} </s>"
        prompt += f"<s>[INST] {user_3.strip()} [/INST]"
    """
    start_token = "<s>"
    end_token = "</s>"
    sys_start = "<<SYS>>"
    sys_end = "<</SYS>>"
    inst_start = "[INST]"
    inst_end = "[/INST]"
    def __init__(self):
        pass

    def __call__(self, messages) -> typing.Tuple[str, typing.Set[str]]:
        """
        messages are of the form
            messages = [
                {
                    "role": "system",
                    "content": "....",
                },
                {
                    "role": "system",
                    "name": "example_user",
                    "content": "....",
                },
                {
                    "role": "system",
                    "name": "example_assistant",
                    "content": "....",
                },
                {
                    "role": "system",
                    "name": "example_user",
                    "content": "....",
                },
                {
                    "role": "system",
                    "name": "example_assistant",
                    "content": "....",
                },
                {
                    "role": "user",
                    "content": "",
                }
            ]
        """
        # Collect all the system messages
        system_messages = []
        user_messages = []
        assistant_messages = []
        role_names = set()
        for message in messages:
            if message["role"] == "system":
                system_messages.append(message)
                if "name" in message:
                    role_names.add(message["name"])
            elif message["role"] == "user":
                user_messages.append(message)
                if "name" in message:
                    role_names.add(message["name"])
            elif message["role"] == "assistant":
                assistant_messages.append(message)
                if "name" in message:
                    role_names.add(message["name"])
            else:
                raise ValueError(f"Unknown role: {message['role']}")
        sys_prompt = self._format_system_messages(system_messages)
        user_prompt = self._format_user_assistant_messages(user_messages, assistant_messages)
        if len(sys_prompt) > 0:
            prompt = \
f"""{sys_prompt}

{user_prompt}"""
        else:
            prompt = user_prompt
        return prompt, role_names

    def _format_system_messages(self, system_message):
        """
        system_message is of the form
        """
        main_system_messages = [sys_msg["content"] for sys_msg in system_message if "name" not in sys_msg]
        main_system_message = "\n".join(main_system_messages)
        example_messages = [(sys_msg["name"], sys_msg["content"]) for sys_msg in system_message if "name" in sys_msg]
        example_messages = [f"`{name}`:\n{msg}" for name, msg in example_messages]
        if len(example_messages) > 0:
            example_message = "\n".join(example_messages)
        else:
            example_message = ""
        if len(example_messages) > 0:
            system_message = \
f"""{main_system_message}

An example of user and assistant interaction is as follows:
{example_message}"""
        else:
            system_message = main_system_message
        if len(system_message) > 0:
            system_prompt = \
    f"""{Llama2ChatFormat.start_token}{Llama2ChatFormat.inst_start} {Llama2ChatFormat.sys_start}
{system_message}
{Llama2ChatFormat.sys_end}"""
        else:
            system_prompt = ""
        return system_prompt
    
    def _format_user_assistant_messages(self, user_messages, assistant_messages):
        """
        user_messages is of the form
        """
        user_messages = [user_msg["content"] for user_msg in user_messages]
        user_messages = [f"{user_msg} {Llama2ChatFormat.inst_end} " for user_msg in user_messages]
        assistant_messages = [assistant_msg["content"] for assistant_msg in assistant_messages]
        if len(assistant_messages) >= len(user_messages):
            assistant_messages = [f"{assistant_msg} {Llama2ChatFormat.end_token}{Llama2ChatFormat.start_token}{Llama2ChatFormat.inst_start} " for assistant_msg in assistant_messages[:-1]] + [f"{assistant_messages[-1]} {Llama2ChatFormat.end_token}"]
        else:
            assistant_messages = [f"{assistant_msg} {Llama2ChatFormat.end_token}{Llama2ChatFormat.start_token}{Llama2ChatFormat.inst_start} " for assistant_msg in assistant_messages]
        # Combine the messages one after the other
        messages = []
        idx = 0
        while idx < len(user_messages) or idx < len(assistant_messages):
            if idx < len(user_messages):
                messages.append(user_messages[idx])
            if idx < len(assistant_messages):
                messages.append(assistant_messages[idx])
            idx += 1
        message = "".join(messages)
        return message
    
if __name__ == "__main__":
    messages = [
        {
            "role": "system",
            "content": "You are a helpful, pattern-following assistant that translates corporate jargon into plain English.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "New synergies will help drive top-line growth.",
        },
        {
            "role": "system",
            "name": "example_assistant",
            "content": "Things working well together will increase revenue.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "Let's circle back when we have more bandwidth to touch base on opportunities for increased leverage.",
        },
        {
            "role": "system",
            "name": "example_assistant",
            "content": "Let's talk later when we're less busy about how to do better.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "This late pivot means we don't have time to boil the ocean for the client deliverable.",
        },
        {
            "role": "system",
            "name": "example_assistant", 
            "content": "Our idea seems to be scooped, don't know how to change direction now."
        },
        {
            "role": "user",
            "content": "We changed the direction of the project, but we don't have time to do it.",
        },
        {
            "role": "assistant",
            "content": "Too many changes do not have time to do it.",
        },
        {
            "role": "user",
            "content": "The pot is boiling, probably the water will spill.",
        }
    ]
    llama2_format_chat = Llama2ChatFormat()
    prompt, role_names = llama2_format_chat(messages)
    print(prompt)
    print('='*50)
    print(role_names)
    print('-'*100)
    messages = [
        {
            "role": "system",
            "content": "You are a helpful, pattern-following assistant that translates corporate jargon into plain English.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "New synergies will help drive top-line growth.",
        },
        {
            "role": "system",
            "name": "example_assistant",
            "content": "Things working well together will increase revenue.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "Let's circle back when we have more bandwidth to touch base on opportunities for increased leverage.",
        },
        {
            "role": "system",
            "name": "example_assistant",
            "content": "Let's talk later when we're less busy about how to do better.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "This late pivot means we don't have time to boil the ocean for the client deliverable.",
        },
        {
            "role": "system",
            "name": "example_assistant", 
            "content": "Our idea seems to be scooped, don't know how to change direction now."
        },
        {
            "role": "user",
            "content": "We changed the direction of the project, but we don't have time to do it.",
        }
    ]
    llama2_format_chat = Llama2ChatFormat()
    prompt, role_names = llama2_format_chat(messages)
    print(prompt)
    print('='*50)
    print(role_names)
    messages = [
        {
            "role": "system",
            "content": "You are a helpful, pattern-following assistant that translates corporate jargon into plain English.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "New synergies will help drive top-line growth.",
        },
        {
            "role": "system",
            "name": "example_assistant",
            "content": "Things working well together will increase revenue.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "Let's circle back when we have more bandwidth to touch base on opportunities for increased leverage.",
        },
        {
            "role": "system",
            "name": "example_assistant",
            "content": "Let's talk later when we're less busy about how to do better.",
        },
        {
            "role": "system",
            "name": "example_user",
            "content": "This late pivot means we don't have time to boil the ocean for the client deliverable.",
        },
        {
            "role": "system",
            "name": "example_assistant", 
            "content": "Our idea seems to be scooped, don't know how to change direction now."
        },
        {
            "role": "user",
            "content": "We changed the direction of the project, but we don't have time to do it.",
        },
        {
            "role": "assistant", 
            "content": "Our idea seems to be scooped, don't know how to change direction now."
        }
    ]
    llama2_format_chat = Llama2ChatFormat()
    prompt, role_names = llama2_format_chat(messages)
    print(prompt)
    print('='*50)
    print(role_names)


import json
import logging
import typing
import requests
from concurrent.futures import ThreadPoolExecutor, as_completed
from model import Model, GenerationResults, GenerationResult
from tgi_server import TGIServer

class TgiModel(Model):
    """
    Wrapper for interacting with models hosted on TGI servers.
    """

    def __init__(self, name: str, log_folder: str = None, **kwargs):
        """
        Initialize TGI Model instance.

        Args:
            name (str): Name of the model hosted on TGI.
            log_folder (str): Optional folder path for logging.
            **kwargs: Additional arguments for the model or configuration.
        """
        self.port = kwargs.get("port", 8080)
        self.hostname = kwargs.get("hostname", "localhost")
        self.base_url = f"http://{self.hostname}:{self.port}"
        self.model_name = name
        self.is_chat_model = kwargs.get("is_chat_model", False)

        # Setup logging
        self.logger = logging.getLogger(f"TgiModel-{name}")
        log_format = "%(asctime)s - %(levelname)s - %(message)s"
        if log_folder:
            handler = logging.FileHandler(f"{log_folder}/tgi_model.log")
        else:
            handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(log_format))
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        self.logger.info("TgiModel initialized")

    def _send_request(self, endpoint: str, payload: dict) -> dict:
        """
        Utility method to send a request to the TGI server.

        Args:
            endpoint (str): API endpoint (e.g., "/v1/completions").
            payload (dict): Request payload.

        Returns:
            dict: Response data from the TGI server.
        """
        url = f"{self.base_url}{endpoint}"
        try:
            response = requests.post(url, json=payload)
            response.raise_for_status()
            return response.json()
        except requests.RequestException as e:
            self.logger.error("Request to %s failed: %s", url, e)
            raise

    def tokenize(self, inputs: typing.List[typing.Any], kwargs: dict = None):
        """
        Tokenize the input strings using the TGI API.

        Args:
            inputs (list[str]): List of input strings to tokenize.
            kwargs (dict): Additional tokenization parameters.

        Returns:
            list: Tokenized inputs from TGI.
        """
        kwargs = kwargs or {}
        self.logger.info("Tokenizing inputs via TGI API: %s", inputs)

        endpoint = "/v1/chat_tokenize" if self.is_chat_model else "/v1/tokenize"
        if self.is_chat_model:
            supported_params = ["max_tokens", "logprobs", "frequency_penalty", "n", 
                "presence_penalty", "response_format", "seed", "stop", "stream", "stream_options",
                "temperature", "tool_choice", "tool_prompt", "tools", "top_logprobs", "top_p"]
            payload = {
                "messages": [{"role": "user", "content": text} for text in inputs] if self.is_chat_model else None
            } | {k: v for k, v in kwargs.items() if k in supported_params}
        else:
            supported_params = ['adapter_id', 'best_of', 'decoder_input_details', 
            'details', 'do_sample', 'frequency_penalty', 'grammar', 'max_new_tokens', 'repetition_penalty', 
            'return_full_text', 'seed', 'stop', 'temperature', 'top_k', 'top_n_tokens', 
            'top_p', 'truncate', 'typical_p', 'watermark']
            payload = {
                "inputs": inputs,
                "parameters": {k: v for k, v in kwargs.items() if k in supported_params}
            }

        return self._send_request(endpoint, payload)

    def _generate_request(self, inputs: typing.Union[str, typing.List[typing.Dict]] , **kwargs) -> typing.List[str]:
        """
        Helper method to send a generation request for a single input.

        Args:
            inputs (str): The input text.
            **kwargs: Additional generation parameters.

        Returns:
            list[str]: Generated text responses.
        """
        endpoint = "/v1/chat/completions" if self.is_chat_model else "/v1/completions"

        # Construct the payload
        payload = {
            "max_tokens": kwargs.get("max_new_tokens", 50),
            "n": kwargs.get("n", 1),
            "temperature": kwargs.get("temperature", 1.0),
            "top_p": kwargs.get("top_p", 0.95),
            "frequency_penalty": kwargs.get("frequency_penalty", 0.0),
            "repetition_penalty": kwargs.get("repetition_penalty", 1.0),
            "stream": False, # Disable streaming for now
        }

        if self.is_chat_model:
            assert isinstance(inputs, typing.List[typing.Dict])
            assert all("role" in msg and "content" in msg for msg in inputs)
            payload["messages"] = inputs
        else:
            assert isinstance(inputs, str)
            payload["prompt"] = inputs

        # Add optional parameters only if present
        optional_params = ["stop", "suffix", "seed"]
        for param in optional_params:
            if param in kwargs:
                payload[param] = kwargs[param]

        # Log the payload
        self.logger.debug("Sending request to %s with payload: %s", endpoint, json.dumps(payload, indent=2))

        # Send request and handle response
        try:
            response = self._send_request(endpoint, payload)
            self.logger.debug("Response from %s: %s", endpoint, json.dumps(response, indent=2))

            # Extract responses based on the model type
            if self.is_chat_model:
                return [{"generation": {choice.get("message", {})}, "finish_reason": choice.get("finish_reason", None)} for choice in response.get("choices", [])]
            else:
                return [{"generation": choice.get("text", ""), "finish_reason": choice.get("finish_reason", None)} for choice in response.get("choices", [])]

        except Exception as e:
            self.logger.error("Error during generation request: %s", e)
            return []

    def generate(self, inputs: typing.Union[str, typing.List[str], typing.List[typing.List[typing.Dict]], typing.List[typing.Dict]], **kwargs) -> GenerationResults:
        """
        Generate text completions for a batch of inputs using the TGI model.

        Args:
            inputs (Union[list[Any], Any]): Input text or list of texts.
            **kwargs: Additional generation parameters.

        Returns:
            GenerationResults: Generated text responses encapsulated in a GenerationResults object.
        """
        if isinstance(inputs, str):
            inputs = [inputs]
        elif not isinstance(inputs, list):
            raise ValueError("Input must be a string or list of strings")
        if not self.is_chat_model:
            assert all(isinstance(input_text, str) for input_text in inputs)
        else:
            if all(isinstance(input_text, dict) for input_text in inputs):
                inputs = [inputs]
            assert all(isinstance(input_text, list) for input_text in inputs)
            assert all(all(isinstance(msg, dict) for msg in input_text) for input_text in inputs)

        self.logger.debug("Generating text for batch of inputs: %s", inputs)
        results = []
        with ThreadPoolExecutor() as executor:
            future_to_input = {}
            for inp_idx, inp in enumerate(inputs):
                num_return_sequences = kwargs.get("num_return_sequences", 1)
                for _ in range(num_return_sequences):
                    future_to_input[executor.submit(self._generate_request, inp, **kwargs)] = inp_idx
            responses = {}
            for future in as_completed(future_to_input):
                inp_idx = future_to_input[future]
                input_text = inputs[inp_idx]
                resp = responses.get(inp_idx, [])
                try:
                    resp.extend(future.result())
                    responses[inp_idx] = resp
                except Exception as e:
                    self.logger.error("Error processing input '%s': %s", input_text, e)
                    responses[inp_idx] = []
            results = []
            for inp_idx, generated_texts in responses.items():
                input_text = inputs[inp_idx]
                generated_text_list = [gen["generation"] for gen in generated_texts]
                finish_reasons = [gen["finish_reason"] for gen in generated_texts]
                results.append(GenerationResult(input_text=input_text, generated_text=generated_text_list, finish_reason=finish_reasons))
        self.logger.debug("Generated batch responses: %s", results)
        return GenerationResults(results=results)

if __name__ == "__main__":
    import os
    model_path = '<root>/models/proof-wala-codet5-small-coq-lean-4-2048/checkpoint-337500'
    if "<root>" in model_path:
        model_path = model_path.replace("<root>/", os.environ.get("ROOT", "").trim('/') + "/")
    import time
    with TGIServer(
        path_to_local_model=model_path,
        port=8080,
        gpus=True,
        logdir=".logs/tgi_server",
        cuda_visible_devices="0,1,2,3,4,5,6,7"
    ) as tgi_server:
        if tgi_server.is_alive():
            tgi_server.ping_test()

            model = TgiModel(
                name="proof-wala-codet5-small-coq-lean",
                hostname=tgi_server.hostname,
                port=tgi_server.port,
                is_chat_model=False
            )
            avg_time = 0
            total_generations = 0
            for run in range(10):
                # Simple test for generation
                inputs = ["a + e = a", "a + 0 = a", "a + 1 = 1 + a", "a*a + 2*a + 1 = (a + 1)*(a + 1)"]
                num_return_sequences = 32
                print("\nGeneration Test:")
                time_start = time.time()
                responses = model.generate(inputs, max_new_tokens=150, temperature=0.75, num_return_sequences=num_return_sequences, stop=["[END]"])
                end_time = time.time()
                for result in responses.results:
                    print(f"Input: {result.input_text}\nResponses: {result.generated_text}\n Finish reasons: {result.finish_reason}")
                total_generations += len(inputs) * num_return_sequences
                print(f"Time taken: {end_time - time_start:.2f} seconds")
                print(f"Total generations: {total_generations}")
                print(f"Average time per generation: {(end_time - time_start) / total_generations:.2f} seconds")
                print('=' * 80)
                avg_time = (avg_time * run + (end_time - time_start)) / (run + 1)
            print(f"Average time taken: {avg_time:.2f} seconds")
            print(f"Avg generations per second: {avg_time / total_generations:.2f}")
        else:
            print("TGI server failed to start")



import typing
import logging
import json
from graphviz import Digraph
from abc import ABC, abstractmethod

def escape_string(string: str):
    escaped_string = string.replace('\n', '\\n')
    return escaped_string

class Edge:
    def __init__(self, label: str, score: float = 0, other_data: typing.Any = None):
        assert isinstance(label, str), "Edge label must be a string"
        assert isinstance(score, (int, float)), "Edge score must be a number"
        self.label = label
        self.score = float(score)
        self.other_data = other_data
        self.equivalent_edges : typing.List['Edge'] = []
    
    def add_equivalent_edge(self, edge: 'Edge'):
        self.equivalent_edges.append(edge)

    def __eq__(self, other):
        if not isinstance(other, Edge):
            return False
        return self.label == other.label
    
    def __hash__(self):
        return hash(self.label)
    
    def __str__(self) -> str:
        return f"Edge({self.label}, {self.score})"
    
    def __repr__(self) -> str:
        return self.__str__()

class Node:
    def __init__(self, name: str, score: float = 0, other_data: typing.Any = None):
        assert isinstance(name, str), "Node name must be a string"
        assert isinstance(score, (int, float)), "Node score must be a number"
        self.name = name
        self.score = float(score)
        self.distance_from_root = float('inf')
        self.cummulative_score = float('inf')
        self.other_data = other_data
        # It is very important to make sure that the members below are lists
        # and not sets, because hash will fail when this objects gets pickled
        # and sent to another process while using multiprocessing
        self.parents : typing.List['Node'] = []  # Keep track of parents for tree representation
        self.children : typing.List['Node'] = []  # Keep track of children for tree representation
        self.edges : typing.List[Edge] = []  # Keep track of edges for graph representation

    def add_child(self, child: 'Node', edge: Edge = None):
        """
        Note: Do not add the same child node to the children list more than once.
        We cannot check for duplicates here because it would be too slow.
        """
        self.children.append(child)
        child.parents.append(self)
        self.edges.append(edge)
    
    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        return self.name == other.name
    
    def __hash__(self):
        return hash(self.name)
    
    def __str__(self) -> str:
        return f"Node({self.name}, {self.score})"
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def __lt__(self, other):
        assert isinstance(other, Node), "Can only compare Node objects"
        return self.score < other.score
    
    def __le__(self, other):
        assert isinstance(other, Node), "Can only compare Node objects"
        return self.score <= other.score
    
    def __gt__(self, other):
        assert isinstance(other, Node), "Can only compare Node objects"
        return self.score > other.score
    
    def __ge__(self, other):
        assert isinstance(other, Node), "Can only compare Node objects"
        return self.score >= other.score

def serialize_edge(edge: 'Edge', visited_edges: set) -> dict:
    if edge in visited_edges:
        return {'label': edge.label}  # Simplified representation for visited edges

    visited_edges.add(edge)

    edge_dict = {
        'label': edge.label,
        'score': edge.score,
        'other_data': edge.other_data,
        'equivalent_edges': [serialize_edge(eq_edge, visited_edges) for eq_edge in edge.equivalent_edges]
    }
    return edge_dict

def serialize_node(node: 'Node', visited_nodes: set, visited_edges: set) -> dict:
    if node in visited_nodes:
        return {'name': node.name}  # Simplified representation for visited nodes

    visited_nodes.add(node)

    node_dict = {
        'name': node.name,
        'score': node.score,
        'distance_from_root': node.distance_from_root,
        'cummulative_score': node.cummulative_score,
        'other_data': node.other_data,
        'children': [serialize_node(child, visited_nodes, visited_edges) for child in node.children],
        'edges': [serialize_edge(edge, visited_edges) for edge in node.edges]
    }
    return node_dict

def serialize_graph(root: 'Node') -> str:
    visited_nodes = set()
    visited_edges = set()
    graph_dict = serialize_node(root, visited_nodes, visited_edges)
    return json.dumps(graph_dict, indent=2)

def deserialize_edge(edge_dict: dict, edges_cache: dict) -> 'Edge':
    edge_label = edge_dict['label']
    
    if edge_label in edges_cache:
        return edges_cache[edge_label]
    
    edge = Edge(
        label=edge_dict['label'],
        score=edge_dict.get('score', 0),
        other_data=edge_dict.get('other_data', None)
    )

    edges_cache[edge_label] = edge

    for eq_edge_dict in edge_dict.get('equivalent_edges', []):
        eq_edge = deserialize_edge(eq_edge_dict, edges_cache)
        edge.add_equivalent_edge(eq_edge)
    
    return edge

def deserialize_node(node_dict: dict, nodes_cache: dict, edges_cache: dict) -> 'Node':
    node_name = node_dict['name']
    
    if node_name in nodes_cache:
        return nodes_cache[node_name]
    
    node = Node(
        name=node_dict['name'],
        score=node_dict.get('score', 0),
        other_data=node_dict.get('other_data', None)
    )
    node.distance_from_root = node_dict.get('distance_from_root', float('inf'))
    node.cummulative_score = node_dict.get('cummulative_score', float('inf'))

    nodes_cache[node_name] = node

    for child_dict in node_dict['children']:
        child_node = deserialize_node(child_dict, nodes_cache, edges_cache)
        node.add_child(child_node)

    for edge_dict in node_dict['edges']:
        edge = deserialize_edge(edge_dict, edges_cache)
        node.edges.append(edge)
    
    return node

def deserialize_graph(json_str: str) -> 'Node':
    graph_dict = json.loads(json_str)
    nodes_cache = {}
    edges_cache = {}
    return deserialize_node(graph_dict, nodes_cache, edges_cache)


class SearchAlgorithm(ABC):
    def __init__(self):
        pass

    def reconstruct_path(self, start_node: Node, goal_node: Node) -> typing.List[Node]:
        """
        Iteratively reconstructs a path from start_node to goal_node, avoiding cycles.
        
        :param start_node: The starting node.
        :param goal_node: The goal node.
        :return: A list of nodes representing the path, or an empty list if no path exists.
        """
        if start_node == goal_node:
            return [start_node]

        visited = set()
        stack = [(start_node, [start_node])]  # Stack holds tuples of (current_node, path_to_node)

        while stack:
            current_node, path = stack.pop()
            visited.add(current_node)

            for child in current_node.children:
                if child in visited:
                    continue  # Skip visited nodes to avoid cycles
                
                if child == goal_node:
                    return path + [child]  # Return the path including the goal node
                
                stack.append((child, path + [child]))

        return []  # No path found if we reach this point
    
    def reconstruct_all_paths(self, start_node: Node, goal_node: Node) -> typing.List[typing.List[Node]]:
        """
        Iteratively reconstructs all paths from start_node to goal_node, handling cycles.
        
        :param start_node: The starting node for paths.
        :param goal_node: The goal node to reach.
        :return: A list of lists, where each inner list represents a path of nodes from start to goal.
        """
        all_paths = []
        queue = [(start_node, [start_node])]  # Queue holds tuples of (current_node, path_to_node)
        
        while queue:
            current_node, path = queue.pop(0)
            
            # If the current node is the goal, add the current path to the list of all paths
            if current_node == goal_node:
                all_paths.append(path)
                continue
            
            # Explore children, avoiding cycles by checking if the child is already in the current path
            for child in current_node.children:
                if child not in path:  # Check to avoid revisiting nodes in the current path (simple cycle avoidance)
                    new_path = path + [child]
                    queue.append((child, new_path))

        return all_paths

    def visualize_search(self, 
        root: Node, 
        save_to_file: str = None, 
        show: bool = True,
        mark_paths: typing.List[typing.List[Node]] = []) -> Digraph:
        """
        Visualizes the tree structure starting from the given root node using Graphviz.
        
        :param root: The root node of the tree to visualize.
        """
        dot = Digraph()
        node_queue = [root]
        visited = set()
        edges = set()
        text_width = 75
        node_num = 0
        unique_node_names = {}
        nodes_in_paths = set()
        edges_in_paths = set()
        for path in mark_paths:
            for i, node in enumerate(path):
                nodes_in_paths.add(node.name)
                if i < len(path) - 1:
                    edges_in_paths.add((node.name, path[i + 1].name))
        while node_queue:
            current_node = node_queue.pop(0)
            full_node_name = current_node.name
            node_name = (current_node.name[:text_width] + "...") if len(current_node.name) > text_width else current_node.name
            node_name += f"\n${current_node.cummulative_score:.2f}$"
            node_name = escape_string(node_name)
            if full_node_name not in unique_node_names:
                unique_node_names[full_node_name] = node_num
                node_num += 1
            full_node_num = unique_node_names[full_node_name]
            if full_node_name in nodes_in_paths:
                dot.node(str(full_node_num), label=node_name, style='filled', fillcolor='lightblue')
            else:
                dot.node(str(full_node_num), label=node_name)
            
            for child_idx, child in enumerate(current_node.children):
                # Add edge from parent to child
                if (current_node, child) not in edges:
                    edges.add((current_node, child))
                    child_name = child.name
                    if child_name not in unique_node_names:
                        unique_node_names[child_name] = node_num
                        node_num += 1
                    child_num = unique_node_names[child_name]
                    edg = current_node.edges[child_idx]
                    edge_labels = []
                    if edg is not None:
                        eqv_edges = edg.equivalent_edges + [edg]
                        for eq_edge in eqv_edges:
                            edge_label = eq_edge.label
                            edge_label = edge_label[:text_width] + "..." if len(edge_label) > text_width else edge_label
                            edge_label += f"\n${eq_edge.score:.2f}$"
                            edge_label = escape_string(edge_label)
                            edge_labels.append(edge_label)
                    else:
                        edge_labels = [None]
                    for edge_label in edge_labels:
                        if (current_node.name, child.name) in edges_in_paths:
                            dot.edge(str(full_node_num), str(child_num), label=edge_label, color='red', penwidth='2.0')
                        else:
                            dot.edge(str(full_node_num), str(child_num), label=edge_label)
                if child not in visited:
                    # Add child to the queue to process its children later
                    node_queue.append(child)
            visited.add(current_node)
        
        # Render the graph to a file (e.g., 'tree.png')
        if save_to_file:
            dot.render(save_to_file, format='png', quiet=True)
        if show:
            dot.view()
        return dot

    @abstractmethod
    def search(
            start: Node, 
            goal: Node,
            heuristic: typing.Callable[[Node, Edge, Node], float], 
            generate_children: typing.Callable[[Node, float], typing.Tuple[typing.List[Node], typing.List[Edge]]] = None,
            parallel_count: int = None,
            build_tree: bool = True,
            timeout_in_secs: float = None,
            logger: logging.Logger = None) -> typing.Tuple[Node, bool, float]:
        """
        Abstract method for implementing search algorithms.
        :param start: The start node for the search.
        :param goal: The goal node for the search.
        :param heuristic: A function that takes a Node and returns a float heuristic value for that node.
        It should return min value for the node which should be explored first.
        :param generate_children: A function that takes a Node and returns a list of child nodes, or None if build_tree is False.
        :param parallel_count: An optional integer for the number of parallel processes to use.
        :param build_tree: A boolean indicating whether to build a tree while searching, default True. 
        Set to False only if the tree is already built and children are already added to nodes.
        :param timeout_in_secs: An optional float for the maximum time to run the search.
        :param logger: An optional logger for logging messages.


        :return: Returns a tuple of the form (Node, bool, float), 
        where the first element is the goal node found in the tree (or the start node if not found),
        the second element is a boolean indicating whether the goal was found, and the third element is the time elapsed.
        """
        pass


#!/usr/bin/env python3

import os
if __name__ == "__main__":
    os.environ["CUDA_VISIBLE_DEVICES"] = "5"
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
import logging
import time
import copy
import numpy as np
from itp_interface.rl.simple_proof_env import ProofState, ProofAction, ProofEnvInfo, ProgressState, ProofTree
from proof_wala.proof_search.search_driver import ProofActionGenerator, ProofStateInfo, ProofSearhHeuristic, Node, Edge
from proof_wala.llm_helpers.model import GenerationResults, Model
from proof_wala.itp.codet5_training_data_formatter import CodeT5TrainingDataset, CoqGptResponse, CoqGptRequest

def get_qed_for_language(language: ProofAction.Language):
    if language == ProofAction.Language.COQ:
        return "Qed."
    elif language == ProofAction.Language.LEAN:
        return "end"
    elif language == ProofAction.Language.LEAN4:
        return "done"
    else:
        raise ValueError(f"Language {language} not supported")

class LlmProofActionGenerator(ProofActionGenerator):
    def __init__(self, 
        model: Model,
        prompt_formatter: typing.Callable[[ProofStateInfo], str],
        response_parser: typing.Callable[[str], typing.List[str]],
        width: int = 4,
        logger: logging.Logger = None,
        **generation_args):
        assert model is not None, "Model cannot be None"
        super().__init__(width)
        self.model = model
        self.prompt_formatter = prompt_formatter
        self.response_parser = response_parser
        self._generation_args = generation_args if generation_args else {}
        self.logger = logger if logger else logging.getLogger(__name__)
        self._generation_args["num_return_sequences"] = self.width
        pass

    def safe_response_parser(self, response: str) -> typing.List[str]:
        try:
            return self.response_parser(response)
        except Exception as e:
            self.logger.error(f"Error while parsing response:\n {e}\nResponse: \n{response}\n")
            return []

    def get_proof_end_for_language(self, language: ProofAction.Language) -> ProofAction:
        qed = get_qed_for_language(language)
        return ProofAction(ProofAction.ActionType.RUN_TACTIC, language, tactics=[qed])

    def generate_actions(self, state_info: ProofStateInfo, k: int = None) -> typing.List[typing.Tuple[float, ProofAction]]:
        state : ProofState = state_info.proof_state
        done : bool = state_info.done
        if done:
            return [(-100.0, ProofAction(ProofAction.ActionType.EXIT, state.language))]
        elif len(state.training_data_format.start_goals) == 0: # No more goals to prove
                qed = get_qed_for_language(state.language)
                return [(-100.0, ProofAction(ProofAction.ActionType.RUN_TACTIC, state.language, tactics=[qed]))]
        else:
            # generate actions using the model
            prompt = self.prompt_formatter(state_info)
            problem = state.theorem_statement_with_name
            self.logger.info(f"Prompt for [{problem}]:\n{prompt}")
            start_time = time.time()
            num_of_sequences = self._generation_args.get("num_return_sequences", 1)
            generation_args = copy.deepcopy(self._generation_args)
            seq_generated = 0
            cumm_generation_result = GenerationResults()
            max_attempts = 7
            while seq_generated < num_of_sequences and max_attempts > 0:
                try:
                    generation_results = self.model.generate(prompt, **generation_args)
                    assert len(generation_results.results) == 1, "Only one prompt is used"
                    seq_generated += len(generation_results.results[0].generated_text)
                    if len(cumm_generation_result.results) == 0:
                        cumm_generation_result.results.extend(generation_results.results)
                    else:
                        assert cumm_generation_result.results[0].input_text == generation_results.results[0].input_text
                        cumm_generation_result.results[0].generated_text.extend(generation_results.results[0].generated_text)
                        cumm_generation_result.results[0].neg_log_likelihood.extend(generation_results.results[0].neg_log_likelihood)
                except Exception as e:
                    # There can be CUDA OOM errors and other issues
                    self.logger.error(f"Error while generating actions: {e}")
                    generation_args["num_return_sequences"] = max(1, generation_args["num_return_sequences"] // 2)
                max_attempts -= 1
            generation_results = cumm_generation_result
            assert len(generation_results.results) == 1, "Only one prompt is used"
            result = generation_results[0]
            raw_outputs = result.generated_text
            neg_log_probabilties = result.neg_log_likelihood
            raw_output_set = set()
            raw_output_lst= []
            neg_log_probabilties_lst = []
            actual_output_size = len(raw_outputs)
            for output, neg_log_prob in zip(raw_outputs, neg_log_probabilties):
                if output not in raw_output_set:
                    raw_output_lst.append(output)
                    neg_log_probabilties_lst.append(neg_log_prob)
                    raw_output_set.add(output)
            end_time = time.time()
            raw_outputs = raw_output_lst
            neg_log_probabilties = neg_log_probabilties_lst
            arg_max_output = raw_outputs[np.argmin(neg_log_probabilties)]
            self.logger.info(f"Generated {len(raw_outputs)} distinct actions (actual output size={actual_output_size}) in {end_time - start_time} seconds")
            self.logger.info(f"Best generated action: {arg_max_output}")
            tactics_list = [self.safe_response_parser(output) for output in raw_outputs]
            tactics_list = [tactics for tactics in tactics_list if len(tactics) > 0]
            actions = [(neg_log_prob, ProofAction(ProofAction.ActionType.RUN_TACTIC, state.language, tactics=tactics)) for neg_log_prob, tactics in zip(neg_log_probabilties, tactics_list)]
            return actions

class CodeT5PromptFormatter:
    def __init__(self, max_token_in_prompt: int, character_per_token: float, no_steps: bool = False):
        self.max_token_in_prompt = max_token_in_prompt
        self.character_per_token = character_per_token
        self.no_steps = no_steps

    def __call__(self, state_info: ProofStateInfo) -> str:
        state : ProofState = state_info.proof_state
        info : ProofEnvInfo = state_info.info
        success = info is None or (info.progress != ProgressState.FAILED and info.progress != ProgressState.STATE_UNCHANGED)
        proof_tree : ProofTree = state.proof_tree
        actions_till_state : typing.List[ProofAction] = proof_tree.actions
        steps = ['\n'.join(action.kwargs['tactics']) for action in actions_till_state]
        coq_response : CoqGptResponse = CoqGptResponse(
            success=success,
            steps=steps[:-1] if len(steps) > 1 else [],
            last_step=steps[-1] if len(steps) > 0 else None,
            error_message=info.error_message if info is not None else None,
            training_data_format=state.training_data_format
        )
        return CodeT5TrainingDataset.prompt_formatter(coq_response, self.max_token_in_prompt, self.character_per_token, self.no_steps)

class CodeT5ResponseParser:        
    def __call__(self, response: str) -> typing.List[str]:
        gpt_request : CoqGptRequest = CodeT5TrainingDataset.response_parser(response)
        return gpt_request.args

class NegLoglikelihoodMinimizingHeuristic(ProofSearhHeuristic):
    def __init__(self):
        pass

    def __call__(self, parent_node: Node, edge: Edge, node: Node) -> float:
        state_info : ProofStateInfo = node.other_data
        state : ProofState = state_info.proof_state
        if state_info.done or len(state.training_data_format.start_goals) == 0:
            return -100.0 # This is a terminal node
        else:
            if parent_node is not None and edge is not None:
                return min(parent_node.cummulative_score + edge.score, node.cummulative_score)
            elif parent_node is None and edge is not None:
                return edge.score
            else:
                return node.cummulative_score

if __name__ == "__main__":
    from proof_wala.proof_search.test_search_driver import test_search_algorithm
    from proof_wala.search.beam_search import BeamSearch
    from proof_wala.search.best_first_search import BestFirstSearch
    from itp_interface.rl.simple_proof_env import ProofExecutorCallback, ProofEnv
    from itp_interface.tools.log_utils import setup_logger
    from datetime import datetime
    model_path = ".log/run_training/new_model/proof-wala-codet5-small-compcert-2048/best-model-20240228-072825"
    is_seq2seq = True
    max_seq_length = 2048
    character_per_token = 3.6
    model = Model(model_path, is_seq2seq=is_seq2seq)
    prompt_formatter = CodeT5PromptFormatter(max_token_in_prompt=max_seq_length, character_per_token=character_per_token)
    response_parser = CodeT5ResponseParser()
    width = 64
    max_new_tokens=175
    temperature=0.75 # Nucleus sampling
    do_sample=True # Nucleus sampling
    top_k=width # Nucleus sampling
    stop_tokens=["[END]"]
    padding=True
    return_full_text=False
    compute_probabilities=True
    time_stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
    log_folder = f".log/test_search_driver/{time_stamp}"
    os.makedirs(log_folder, exist_ok=True)
    log_file = os.path.join(log_folder, "test_search_driver.log")
    logger = setup_logger("test_search_driver", log_file)
    with model:
        generator = LlmProofActionGenerator(
            model, prompt_formatter, response_parser, width, logger,
            max_new_tokens=max_new_tokens, temperature=temperature, do_sample=do_sample, top_k=top_k,
            stop_tokens=stop_tokens, padding=padding, return_full_text=return_full_text, compute_probabilities=compute_probabilities)
        proof_exec_callback = ProofExecutorCallback(
            project_folder=".",
            file_path="src/proof_wala/data/proofs/coq/simple2/thms.v"
        )
        theorem_names = [
            # "finite_unary_functions",
            "nat_add_comm",
            "double_neg",
            "trival_implication",
            "modus_ponens",
            "modus_tollens",
            "disjunctive_syllogism",
            "contrapositive",
            "nat_zero_add",
            "nat_add_zero",
            "nat_add_succ",
            "nat_succ_add"
        ]
        for search_aglo in [BeamSearch(3)]:
            algo_name = search_aglo.__class__.__name__
            print(f"Running tests for {algo_name}")
            for theorem_name in theorem_names:
                print(f"Trying to prove {theorem_name}")
                language = ProofAction.Language.COQ
                always_retrieve_thms = False
                env = ProofEnv("test", proof_exec_callback, theorem_name, max_proof_depth=100, always_retrieve_thms=always_retrieve_thms, logger=logger)
                test_search_algorithm(
                    exp_name=os.path.join(algo_name, theorem_name),
                    algo=search_aglo,
                    proof_env=env,
                    proof_search_heuristic=NegLoglikelihoodMinimizingHeuristic(),
                    action_generator=generator,
                    search_width=width,
                    attempt_count=30,
                    timeout_in_secs=1200)
                print('-' * 80)
            print('=' * 80)


import os
import json
import typing
from comet_ml import Experiment

class CometHelper(object):
    def __init__(self, secret_file: str = "comet_key.json", secret_folder: str = ".secrets"):
        assert secret_file is not None, "Secret file cannot be None"
        secret_file_path = os.path.join(secret_folder, secret_file)
        assert os.path.exists(secret_file_path), f"Secret file {secret_file} does not exist in the folder {secret_folder}"
        secret = None
        with open(secret_file_path, "r") as f:
            secret = f.read().strip()
            secret = json.loads(secret)
        self.project_name = secret["project_name"]
        self.api_key = secret["api_key"]
        self.workspace = secret["workspace"]
    
    def get_experiment(self, experiment_name: str, tags: typing.List[str] = None) -> Experiment:
        assert experiment_name is not None, "Experiment name cannot be None"
        exp = Experiment(project_name=self.project_name,
                         workspace=self.workspace,
                         api_key=self.api_key)
        exp.set_name(experiment_name)
        if tags is not None:
            exp.add_tags(tags)
        return exp
    

if __name__ == "__main__":
    helper = CometHelper(secret_file="comet_key.json")
    exp = helper.get_experiment(experiment_name="test", tags=["test"])
    exp.log_metric("test", 1)


import os
import torch
import typing
import logging
import json
import random
import shutil
import uuid
import time
from enum import Enum
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from torch.utils.data import Dataset, DataLoader
from transformers import (
    AutoModelForSeq2SeqLM,
    AutoModelForCausalLM,
    AutoTokenizer,
    BitsAndBytesConfig,
    DataCollatorForSeq2Seq,
    RobertaTokenizer,
    StoppingCriteria,
    StoppingCriteriaList,
    TrainingArguments,
    TrainerCallback,
)
from itp_interface.tools.log_utils import setup_logger
from transformers.trainer_callback import TrainerControl, TrainerState
from peft import LoraConfig
from comet_ml import Experiment

try:
    from .huggingface_model_map import get_empty_config, trim_model_kwargs, get_model_class, get_decoder_start_token_id
except ImportError:
    from huggingface_model_map import get_empty_config, trim_model_kwargs, get_model_class, get_decoder_start_token_id

try:
    from .cuda_context import CudaContext
except ImportError:
    from cuda_context import CudaContext

try:
    from .comet_helper import CometHelper
except ImportError:
    from comet_helper import CometHelper

try:
    from .custom_sft_trainer import GenerateEvalSFTTrainer
except ImportError:
    from custom_sft_trainer import GenerateEvalSFTTrainer

try:
    from .custom_s2s_trainer import GenerateEvalS2STrainer
except ImportError:
    from custom_s2s_trainer import GenerateEvalS2STrainer

class AutoRegressiveSequenceSearch(Enum):
    NucleusSampling = "nucleus_sampling"
    BeamSearch = "beam_search"
    def __str__(self):
        return self.value

class TrainingDataFormatterCallback:
    def __init__(self):
        pass
    
    def __call__(self, training_data_example: typing.Any) -> typing.List[str]:
        raise NotImplementedError
    
    def get_prompt_and_completion(self, training_data_example: typing.Any) -> typing.List[typing.Tuple[str, str]]:
        raise NotImplementedError
    
    def get_stopping_token(self):
        raise NotImplementedError

class LogMetricCallback(TrainerCallback):
    def __init__(self, model: "Model"):
        self.model = model
        self._idx = 0
        self._last_log_time = time.time()
        self._last_log_step = 0
        self._steps_per_sec = 1e-6
        self._exp_completion_time = 1e-6
        self._alpha = 0.9

    def on_step_end(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        # Check if we should run evaluation
        if state.global_step == 2 and not self.model._no_init_eval: # Just force evaluation at the beginning to ensure that the model is working
            control.should_log = True
            control.should_evaluate = True
        super().on_step_end(args, state, control, **kwargs)
        return control

    def on_log(self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs):
        if len(state.log_history) > 0:
            time_now = time.time()
            time_elapsed = time_now - self._last_log_time
            steps_elapsed = state.global_step - self._last_log_step
            if time_elapsed == 0:
                time_elapsed = 1e-6
            steps_per_second = steps_elapsed / time_elapsed
            if steps_per_second == 0:
                steps_per_second = 1 / time_elapsed # Pretend that we did 1 step
            expected_completion_time = state.max_steps / steps_per_second
            if self._last_log_step == 0:
                self._steps_per_sec = steps_per_second
                self._exp_completion_time = expected_completion_time
            else:
                self._steps_per_sec = self._alpha * self._steps_per_sec + (1 - self._alpha) * steps_per_second
                self._exp_completion_time = self._alpha * self._exp_completion_time + (1 - self._alpha) * expected_completion_time
            self._last_log_time = time_now
            self._last_log_step = state.global_step
            metrices = state.log_history[self._idx:]
            for metrics in metrices:
                metrics = dict(metrics)
                metrics["steps_per_sec"] = self._steps_per_sec
                metrics["exp_completion_time_in_sec"] = self._exp_completion_time
                metrics["exp_completion_time_in_hours"] = self._exp_completion_time / 3600
                if self.model._should_use_comet:
                    self.model._comet_experiment.log_metrics(metrics, step=state.global_step, epoch=state.epoch)
                metric_json = json.dumps(metrics)
                self.model.metric_logger.info(metric_json)
            self._idx = len(state.log_history)
        super().on_log(args, state, control, **kwargs)

@dataclass_json
@dataclass
class GenerationResult(object):
    input_text: str
    generated_text: typing.List[str] = field(default_factory=list)
    neg_log_likelihood: typing.List[float] = field(default_factory=list)
    finish_reason: typing.List[str] = field(default_factory=list)

@dataclass_json
@dataclass
class GenerationResults(object):
    results: typing.List[GenerationResult] = field(default_factory=list)

    def __iter__(self):
        return iter(self.results)
    
    def __getitem__(self, key):
        assert isinstance(key, int), "Please provide an integer key"
        return self.results[key]

class StopOnTokens(StoppingCriteria):
    """Stopping criteria based on a list of tokens"""
    def __init__(self, stop_tokens: typing.List[str], tokenizer: AutoTokenizer, input_length: int):
        self.stop_token_ids = stop_tokens
        # self.input_length = input_length
        self.tokenizer = tokenizer
        self.stop_tokens = stop_tokens
        self.cuda_context = CudaContext.get_default_context()
        stop_token_ids = [self.tokenizer.encode(stop_word, add_special_tokens=False)
                    for stop_word in self.stop_tokens]
        stop_token_ids = [self.cuda_context.try_get_gpu(torch.LongTensor(x)) for x in stop_token_ids]
        self.max_stop_token_id_length = max([len(x) for x in self.stop_token_ids])
        self.pad_token_tensor = self.cuda_context.try_get_gpu(torch.LongTensor([self.tokenizer.pad_token_id]))
        self.stop_decisions = {}
        self.input_length = input_length

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs) -> bool:
        if input_ids.shape[-1] <= self.input_length and not isinstance(self.tokenizer, RobertaTokenizer):
            return False
        stop_decisions = [-1] * input_ids.shape[0]
        for idx in range(input_ids.shape[0]):
            input_ids_slice = input_ids[idx][-self.max_stop_token_id_length:]
            stop_decisions[idx] = self.stop_decisions.get(idx, -1)
            if len(input_ids_slice.shape) > 0 and stop_decisions[idx] == -1:
                decoded_slice = self.tokenizer.decode(input_ids_slice)
                for stop_token in self.stop_tokens:
                    if decoded_slice.endswith(stop_token):
                        stop_decisions[idx] = len(input_ids[idx])
                        self.stop_decisions[idx] = stop_decisions[idx]
            elif stop_decisions[idx] != -1:
                # Fill the rest of the sequence with the pad token
                # Change input_ids[idx][stop_decisions[idx]:] to the pad token
                for i in range(stop_decisions[idx], len(input_ids[idx])):
                    input_ids[idx][i] = self.pad_token_tensor
        return all([x != -1 for x in stop_decisions])

class Model(object):
    """Wrapper for a Llama model"""

    def __init__(self, name: str, training_args: TrainingArguments = None, log_folder: str = None, **kwargs):
        assert name is not None, "Please provide a model name"
        self.name = name
        self.cuda_context = CudaContext.get_default_context()
        self.training_args = training_args
        gpu_id_local = os.environ.get("LOCAL_RANK", None)
        gpu_id_global = os.environ.get("RANK", None)
        gpu_id_unique = str(uuid.uuid4())
        self.gpu_id = f"{gpu_id_local}-{gpu_id_global}-{gpu_id_unique}"
        code_logger = logging.getLogger(f"ModelCode-{self.gpu_id}") if log_folder is None else setup_logger(f"ModelCode-{self.gpu_id}", f"{log_folder}/model_code-{self.gpu_id}.log")
        metric_format = '{"time": "%(asctime)s", "metrics": %(message)s}'
        metric_logger = logging.getLogger(f"ModelMetrics-{self.gpu_id}") if log_folder is None else setup_logger(f"ModelMetrics-{self.gpu_id}", f"{log_folder}/model_metrics-{self.gpu_id}.jsonl", format=metric_format)
        self.code_logger = code_logger
        self.metric_logger = metric_logger
        self._is_loaded = False
        self._kwargs = kwargs
        self._should_load_model = self._kwargs.get("load_model", True)
        self._should_use_lora = self._kwargs.get("use_lora", True)
        self._comet_experiment_name = self._kwargs.get("comet_experiment", None)
        self._should_use_comet = self._comet_experiment_name is not None
        self._is_seq2seq = self._kwargs.get("is_seq2seq", False)
        self._comet_experiment = None
        self._no_init_eval = self._kwargs.get("no_init_eval", False)
        self._random_init = self._kwargs.get("random_init", False)
        if self._random_init:
            assert "config_class_name" in self._kwargs, "Please provide the config class name for random initialization"
        self._config_class_name = self._kwargs.get("config_class_name", None)
        pass

    def _get_model_args(self):
        return {k: v for k, v in self._kwargs.items() if k in
                ["token", "quantization_config"]
        }

    def _model_init(self) -> typing.Union[AutoModelForSeq2SeqLM, AutoModelForCausalLM]:
        model_args = self._get_model_args()
        if self._random_init:
            assert self._tokenizer is not None, "Tokenizer is not loaded"
            # initialize the model with random weights
            # get the config class from the config class name
            random_config = get_empty_config(self._config_class_name)
            random_config.decoder_start_token_id = get_decoder_start_token_id(self._config_class_name, self._tokenizer)
            model_class = get_model_class(self._config_class_name)
            model_args = trim_model_kwargs(model_args, self._config_class_name)
        if self.is_sequence_2_sequence():
            if self._random_init:
                temp_model = model_class(random_config, **model_args)
            else:
                temp_model = AutoModelForSeq2SeqLM.from_pretrained(self.name, **model_args)
        else:
            if self._random_init:
                temp_model = model_class(random_config, **model_args)
            else:
                temp_model = AutoModelForCausalLM.from_pretrained(self.name, **model_args)
        return temp_model

    def get_tokenizer(self):
        tokenizer_args = {k: v for k, v in self._kwargs.items() if k in
                ["token", "pad_token", "eos_token", "bos_token", "unk_token", "mask_token", "additional_special_tokens", "max_length", "padding", "truncation", "cls_token", "sep_token", "mask_token", "padding_side"]
        }
        tokenizer : AutoTokenizer = AutoTokenizer.from_pretrained(self.name, **tokenizer_args)
        if "llama-2" in self.name.lower():
            if "pad_token" not in tokenizer_args:
                tokenizer_args["pad_token"] = tokenizer.eos_token
            if "padding_side" not in tokenizer_args:
                tokenizer_args["padding_side"] = "left" # This is very import for the stop token logic
            if "cls_token" not in tokenizer_args:
                tokenizer_args["cls_token"] = tokenizer_args["pad_token"]
            if "sep_token" not in tokenizer_args:
                tokenizer_args["sep_token"] = tokenizer_args["pad_token"]
            if "mask_token" not in tokenizer_args:
                tokenizer_args["mask_token"] = tokenizer_args["pad_token"]
        if "pad_token" in tokenizer_args:
            tokenizer.pad_token = tokenizer_args["pad_token"]
        if "eos_token" in tokenizer_args:
            tokenizer.eos_token = tokenizer_args["eos_token"]
        if "bos_token" in tokenizer_args:
            tokenizer.bos_token = tokenizer_args["bos_token"]
        if "unk_token" in tokenizer_args:
            tokenizer.unk_token = tokenizer_args["unk_token"]
        if "additional_special_tokens" in tokenizer_args:
            tokenizer.additional_special_tokens = tokenizer_args["additional_special_tokens"]
        if "cls_token" in tokenizer_args:
            tokenizer.cls_token = tokenizer_args["cls_token"]
        if "sep_token" in tokenizer_args:
            tokenizer.sep_token = tokenizer_args["sep_token"]
        if "mask_token" in tokenizer_args:
            tokenizer.mask_token = tokenizer_args["mask_token"]
        if "padding_side" in tokenizer_args:
            tokenizer.padding_side = tokenizer_args["padding_side"]
        return tokenizer, tokenizer_args

    def load(self):
        if self._is_loaded:
            return
        if self._should_use_lora and "quantization_config" not in self._kwargs:
            quantization_config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_compute_dtype=getattr(torch,"float16"),
                bnb_4bit_use_double_quant=False,
            )
            self._kwargs["quantization_config"] = quantization_config
        quantization_config: BitsAndBytesConfig = self._kwargs.get("quantization_config", None)
        if quantization_config is not None and quantization_config.bnb_4bit_compute_dtype == torch.float16 and quantization_config.load_in_4bit:
            major, _ = torch.cuda.get_device_capability()
            if major >= 8:
                self.code_logger.info("The GPU supports bfloat16: accelerate training with bf16=True")
        if self._should_use_lora and "lor_config" not in self._kwargs:
            lora_config = LoraConfig(
                lora_alpha=16,
                lora_dropout=0.1,
                r=64,
                bias="none",
                task_type="CAUSAL_LM",
            )
            self._kwargs["lora_config"] = lora_config
        self._tokenizer, tokenizer_args = self.get_tokenizer()
        if self._should_load_model:
            if quantization_config is None:
                self._model : typing.Union[AutoModelForSeq2SeqLM, AutoModelForCausalLM] = self.cuda_context.try_get_gpu(self._model_init())
            else:
                self._model = self._model_init()
            if hasattr(self._model, "config"):
                if hasattr(self._model.config, "use_cache"):
                    self._model.config.use_cache = False
                if hasattr(self._model.config, "pretraining_tp"):
                    self._model.config.pretraining_tp = 1
        else:
            self._model = None
        self._is_loaded = True
        self._model_args = self._get_model_args()
        self._tokenizer_args = tokenizer_args


    def __enter__(self):
        self.load()
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        pass

    
    def is_sequence_2_sequence(self):
        return self._is_seq2seq

    def tokenize(self, tokenizer: AutoTokenizer, inputs: typing.List[str], kwargs: dict = None):
        kwargs = kwargs if kwargs is not None else self._kwargs
        tokenizer_args = {k: v for k, v in kwargs.items() if k in 
                ["max_length", "padding", "truncation"]}
        tokenized_input = tokenizer(
            inputs, 
            return_tensors="pt",
            **tokenizer_args)
        return tokenized_input

    def generate(self, inputs: typing.Union[typing.List[str], str], **kwargs) -> GenerationResults:
        assert self._is_loaded, "Model or tokenizer is not loaded"
        assert self._model is not None, "Model is not loaded"
        if isinstance(inputs, str):
            inputs = [inputs]
        stop_tokens_are_present = "stop_tokens" in kwargs
        if "stop_tokens" in kwargs:
            stop_tokens = kwargs["stop_tokens"]
        else:
            stop_tokens = [self._tokenizer.eos_token]
        if "auto_regressive_sequence_search" in kwargs:
            auto_regressive_sequence_search = kwargs["auto_regressive_sequence_search"]
            assert isinstance(auto_regressive_sequence_search, AutoRegressiveSequenceSearch), "Please provide a valid auto_regressive_sequence_search"
        else:
            auto_regressive_sequence_search = AutoRegressiveSequenceSearch.NucleusSampling
        tokenized_input = self.tokenize(self._tokenizer, inputs, kwargs)
        input_ids=self.cuda_context.try_get_gpu(tokenized_input['input_ids'])
        attention_mask=self.cuda_context.try_get_gpu(tokenized_input['attention_mask'])
        max_input_length = input_ids.shape[-1]
        if "llama-2" in self.name.lower() or stop_tokens_are_present:
            stopping_criteria = StopOnTokens(stop_tokens, tokenizer=self._tokenizer, input_length=max_input_length)
            stopping_criteria = StoppingCriteriaList([stopping_criteria])
        else:
            stopping_criteria = None
        if auto_regressive_sequence_search == AutoRegressiveSequenceSearch.NucleusSampling:
            generate_args = {k: v for k, v in kwargs.items() if k in
            ["do_sample", "top_k", "top_p", "num_return_sequences", "temperature", "max_new_tokens"]}
        elif auto_regressive_sequence_search == AutoRegressiveSequenceSearch.BeamSearch:
            generate_args = {k: v for k, v in kwargs.items() if k in
            ["num_beams", "num_return_sequences", "max_new_tokens"]}
        else:
            raise ValueError(f"Invalid auto_regressive_sequence_search: {auto_regressive_sequence_search}")
        if "num_return_sequences" not in generate_args:
            generate_args["num_return_sequences"] = 1
        num_return_sequences = generate_args["num_return_sequences"]
        if "llama-2" in self.name.lower():
            generated_output = self._model.generate(
                input_ids=input_ids, 
                attention_mask=attention_mask,
                stopping_criteria=stopping_criteria,
                eos_token_id=self._tokenizer.eos_token_id,
                return_dict_in_generate=True,
                **generate_args)
        else:
            generated_output = self._model.generate(
                input_ids=input_ids, 
                attention_mask=attention_mask,
                stopping_criteria=stopping_criteria,
                return_dict_in_generate=True,
                **generate_args)
        target = generated_output["sequences"]
        return_full_text = kwargs.get("return_full_text", False)
        if not self.is_sequence_2_sequence() and not return_full_text and "llama-2" in self.name.lower():
            # Replace the target with padding tokens after the input_ids
            target[:, :max_input_length] = self._tokenizer.pad_token_id # This helps in removing the input text from the generated text
        skip_special_tokens = kwargs.get("skip_special_tokens", True)
        decoded_str = self._tokenizer.batch_decode(target, skip_special_tokens=skip_special_tokens)
        idx = 0
        generation_results = GenerationResults()
        compute_probabilities = kwargs.get("compute_probabilities", False)
        neg_log_likelihoods = []
        if compute_probabilities:
            # no gradient computation
            with torch.no_grad():
                for i in range(len(inputs)):
                    inp_i = input_ids[i, :].unsqueeze(0)
                    attention_i = attention_mask[i, :].unsqueeze(0)
                    for j in range(num_return_sequences):
                        target_j = target[i*num_return_sequences+j].unsqueeze(0)
                        out_j = self._model(input_ids=inp_i, attention_mask=attention_i, labels=target_j)
                        neg_log_likelihoods.append(float(out_j.loss.detach().cpu().numpy()))
        for text in inputs:
            generated_text = decoded_str[idx: idx + num_return_sequences]
            if compute_probabilities:
                probabilties = neg_log_likelihoods[idx: idx + num_return_sequences]
            else:
                probabilties = [0.0 for _ in range(num_return_sequences)]
            idx += num_return_sequences
            result = GenerationResult(input_text=text, generated_text=generated_text, neg_log_likelihood=probabilties)
            generation_results.results.append(result)
        return generation_results
   
    def _generate_completion_callback(self, formatter_callback: TrainingDataFormatterCallback):
        def _callback(dataset: Dataset):
            prompts_and_completions = formatter_callback.get_prompt_and_completion(dataset)
            batch_prompts = [x[0] for x in prompts_and_completions]
            batch_completions = [x[1] for x in prompts_and_completions]
            batch_completions_tokenized = self._tokenizer(batch_completions, return_tensors="pt", padding=True, return_length=True)
            max_new_tokens = max(batch_completions_tokenized["length"])
            stop_tokens = [formatter_callback.get_stopping_token(), self._tokenizer.eos_token]
            # Remove the stop tokens which empty
            stop_tokens = [x for x in stop_tokens if x is not None and isinstance(x, str) and x != ""]
            # Now generate all the completions
            generated_completions = self.generate(batch_prompts,
                            max_new_tokens=max_new_tokens,
                            temperature=0.1, # Nucleus sampling
                            do_sample=True, # Nucleus sampling
                            top_k=5, # Nucleus sampling
                            # num_beams=5, # Beam search
                            num_return_sequences=1,
                            stop_tokens=stop_tokens,
                            padding=True,
                            #truncation=True,
                            return_full_text=False)
            # Now compare the completions
            labels = batch_completions
            preds = [gen_res.generated_text for gen_res in generated_completions]
            return batch_prompts, labels, preds
        return _callback

    def _exact_match_metric_callback(self, datasets, formatter_callback: TrainingDataFormatterCallback, completion_callback: typing.Callable[[Dataset], typing.Tuple[typing.List[str], typing.List[str], typing.List[str]]]):
        actual_eval_datasets = datasets 
        def _callback(batch_size: int, trainer_state: TrainerState):
            nonlocal actual_eval_datasets
            if actual_eval_datasets is None:
                return {}
            if not isinstance(actual_eval_datasets, dict):
                eval_datasets = {"eval_dataset": actual_eval_datasets}
            avg = {
                'exact_match': 0.0
            }
            for eval_dataset_name in eval_datasets:
                eval_dataset = eval_datasets[eval_dataset_name]
                # create a dataloader with batch_size
                dataloader = DataLoader(eval_dataset, batch_size=batch_size, shuffle=False)
                all_prompts = []
                all_completions = []
                all_predictions = []
                correct_pred_idx = []
                for batch_idx, dataset in enumerate(dataloader):
                    prompts_and_completions = formatter_callback.get_prompt_and_completion(dataset)
                    batch_prompts = [x[0] for x in prompts_and_completions]
                    batch_completions = [x[1] for x in prompts_and_completions]
                    all_prompts.extend(batch_prompts)
                    all_completions.extend(batch_completions)
                    batch_completions_tokenized = self._tokenizer(batch_completions, return_tensors="pt", padding=True, return_length=True)
                    max_new_tokens = max(batch_completions_tokenized["length"])
                    # Now generate all the completions
                    generated_completions = self.generate(batch_prompts,
                                    max_new_tokens=max_new_tokens,
                                    temperature=0.1, # Nucleus sampling
                                    do_sample=True, # Nucleus sampling
                                    top_k=5, # Nucleus sampling
                                    # num_beams=5, # Beam search
                                    num_return_sequences=1,
                                    stop_tokens=[formatter_callback.get_stopping_token(), self._tokenizer.eos_token],
                                    padding=True,
                                    #truncation=True,
                                    return_full_text=False)
                    # Now compare the completions
                    labels = batch_completions
                    preds = [gen_res.generated_text[0] for gen_res in generated_completions]
                    all_predictions.extend(preds)
                    for _idx, (label, pred) in enumerate(zip(labels, preds)):
                        if label == pred:
                            correct_pred_idx.append(batch_idx * batch_size + _idx)
                # Calculate Exact Match (EM)
                wrong_pred_idx = list(set(range(len(all_completions))) - set(correct_pred_idx))
                em = len(correct_pred_idx) / len(all_completions)
                avg['exact_match'] += em
                prefix = f"[Step = {trainer_state.global_step}] [Epoch = {trainer_state.epoch}] [Dataset = {eval_dataset_name}]"
                self.code_logger.info(f"{prefix} [EM = {em}]")
                # Sample some examples and log them
                min_num_examples = 20
                sampled_correct_pred_idx = random.sample(correct_pred_idx, min(min_num_examples, len(correct_pred_idx)))
                for _idx in sampled_correct_pred_idx:
                    self.code_logger.info(f"{prefix} Prompt [{_idx + 1}]: {all_prompts[_idx]}")
                    self.code_logger.info(f"{prefix} Label [{_idx + 1}]: {all_completions[_idx]}")
                    self.code_logger.info(f"{prefix} Correct Prediction [{_idx + 1}]: {all_predictions[_idx]}")
                    if self._comet_experiment:
                        full_text = f"{prefix}\nPrompt:\n{all_prompts[_idx]}\nLabel:\n{all_completions[_idx]}\nCorrect Prediction:\n{all_predictions[_idx]}"
                        self._comet_experiment.log_text(full_text, step=trainer_state.global_step, metadata={"type": "correct_prediction"})
                sampled_wrong_pred_idx = random.sample(wrong_pred_idx, min(min_num_examples, len(wrong_pred_idx)))
                for _idx in sampled_wrong_pred_idx:
                    self.code_logger.info(f"{prefix} Prompt [{_idx + 1}]: {all_prompts[_idx]}")
                    self.code_logger.info(f"{prefix} Label [{_idx + 1}]: {all_completions[_idx]}")
                    self.code_logger.info(f"{prefix} Wrong Prediction [{_idx + 1}]: {all_predictions[_idx]}")
                    if self._comet_experiment:
                        full_text = f"{prefix}\nPrompt:\n{all_prompts[_idx]}\nLabel:\n{all_completions[_idx]}\nWrong Prediction:\n{all_predictions[_idx]}"
                        self._comet_experiment.log_text(full_text, step=trainer_state.global_step, metadata={"type": "wrong_prediction"})
            avg['exact_match'] /= len(eval_datasets)
            return avg
        return _callback
    
    def _generative_compute_metrics(self):
        def _compute_metrics_callback(
                eval_dataset_name: str,
                trainer_state: TrainerState,
                prompts: typing.List[str], 
                labels: typing.List[str], 
                completions: typing.List[typing.List[str]]):
                avg = {'exact_match': 0.0}
                correct_pred_idx = []
                for _idx, (label, completion) in enumerate(zip(labels, completions)):
                    for _completion in completion:
                        if label == _completion:
                            correct_pred_idx.append(_idx)
                            break
                wrong_pred_idx = list(set(range(len(labels))) - set(correct_pred_idx))
                em = len(correct_pred_idx) / len(labels)
                avg['exact_match'] += em
                prefix = f"[Step = {trainer_state.global_step}] [Epoch = {trainer_state.epoch}] [Dataset = {eval_dataset_name}]"
                self.code_logger.info(f"{prefix} [EM = {em}]")
                # Sample some examples and log them
                min_num_examples = 15
                sampled_correct_pred_idx = random.sample(correct_pred_idx, min(min_num_examples, len(correct_pred_idx)))
                for _idx in sampled_correct_pred_idx:
                    self.code_logger.info(f"{prefix} Prompt [{_idx + 1}]: {prompts[_idx]}")
                    self.code_logger.info(f"{prefix} Label [{_idx + 1}]: {labels[_idx]}")
                    self.code_logger.info(f"{prefix} Correct Prediction [{_idx + 1}]: {completions[_idx]}")
                    if self._comet_experiment:
                        full_text = f"{prefix}\nPrompt:\n{prompts[_idx]}\nLabel:\n{labels[_idx]}\nCorrect Prediction:\n{completions[_idx]}"
                        self._comet_experiment.log_text(full_text, step=trainer_state.global_step, metadata={"type": "correct_prediction"})
                sampled_wrong_pred_idx = random.sample(wrong_pred_idx, min(min_num_examples, len(wrong_pred_idx)))
                for _idx in sampled_wrong_pred_idx:
                    self.code_logger.info(f"{prefix} Prompt [{_idx + 1}]: {prompts[_idx]}")
                    self.code_logger.info(f"{prefix} Label [{_idx + 1}]: {labels[_idx]}")
                    self.code_logger.info(f"{prefix} Wrong Prediction [{_idx + 1}]: {completions[_idx]}")
                    if self._comet_experiment:
                        full_text = f"{prefix}\nPrompt:\n{prompts[_idx]}\nLabel:\n{labels[_idx]}\nWrong Prediction:\n{completions[_idx]}"
                        self._comet_experiment.log_text(full_text, step=trainer_state.global_step, metadata={"type": "wrong_prediction"})
                return avg
        return _compute_metrics_callback
    
    def _log_metric_callback(self) -> TrainerCallback:
        return LogMetricCallback(self)

    def _seq2seq_tokenizer(self, dataset: Dataset, training_data_formatter_callback: TrainingDataFormatterCallback, max_seq_length: int = 512):
        prompts_and_completions = training_data_formatter_callback.get_prompt_and_completion(dataset)
        inputs = [x[0] for x in prompts_and_completions]
        labels = [x[1] for x in prompts_and_completions]
        tokenizer_args = {k: v for k, v in self._tokenizer_args.items() if k in 
        ["max_length", "padding", "truncation"]}
        model_inputs = self._tokenizer(inputs, return_tensors="pt", **tokenizer_args)
        label_args = {k: v for k, v in self._tokenizer_args.items() if k in
        ["padding", "truncation"]}
        label_args["max_length"] = max_seq_length
        model_labels = self._tokenizer(labels, return_tensors="pt", **label_args)
        model_inputs["labels"] = model_labels["input_ids"]
        return model_inputs

    def train(self,
            time_str: str,
            training_data_formatter_callback: TrainingDataFormatterCallback,
            train_dataset: Dataset, 
            eval_dataset: typing.Union[Dataset, typing.Dict[str, Dataset]] = None, 
            compute_metrics: typing.Optional[typing.Callable[[str, TrainerState, typing.List[str], typing.List[str], typing.List[typing.List[str]]], typing.Dict]] = None,
            callbacks: typing.List[TrainerCallback] = None):
        assert self._is_loaded, "Model or tokenizer is not loaded"
        assert self.training_args is not None, "Please provide training arguments"
        eval_dataset_is_dict = isinstance(eval_dataset, dict)
        if compute_metrics is None:
            compute_metrics = self._generative_compute_metrics()
            # paramless_compute_metrics = self._exact_match_metric_callback(eval_dataset, training_data_formatter_callback)
        if callbacks is None:
            callbacks = []
        # unload the model from the GPU
        self.cuda_context.free_cuda_cache_if_possible()
        lora_config = self._kwargs.get("lora_config", None)
        max_seq_length = self._kwargs.get("max_seq_length", None)
        if self._should_use_comet:
            self._comet_helper = CometHelper()
            self._comet_experiment : Experiment = self._comet_helper.get_experiment(self._comet_experiment_name)
            # Check if the experiment exists
            self._comet_experiment.log_parameters(self._kwargs)
        if self.is_sequence_2_sequence():
            data_collator = DataCollatorForSeq2Seq(tokenizer=self._tokenizer, model=self._model)
            trainer = GenerateEvalS2STrainer(
                model=(self.name if self._model is None else self._model), # This will anyway create a new model
                args=self.training_args,
                data_collator=data_collator,
                train_dataset=train_dataset,
                eval_dataset=eval_dataset,
                tokenizer=self._tokenizer,
                callbacks=callbacks,
                preprocess_logits_for_metrics=(lambda _x, _y: fake_logit_tensor), # no need to compute logits and reduce memory usage
                tokenizer_func=lambda x: self._seq2seq_tokenizer(x, training_data_formatter_callback, max_seq_length=max_seq_length),
                generate_callback=self._generate_completion_callback(training_data_formatter_callback),
                generative_compute_metrics=compute_metrics,
                logger=self.code_logger,
                model_init=self._model_init if self._random_init else None,
            )
        else:
            fake_logit_tensor = torch.tensor([0.0])
            trainer = GenerateEvalSFTTrainer(
                model=(self.name if self._model is None else self._model), # This will anyway create a new model
                args=self.training_args,
                # data_collator=self._collate_fn,
                train_dataset=train_dataset,
                eval_dataset=eval_dataset,
                tokenizer=self._tokenizer,
                callbacks=callbacks,
                # dataset_text_field="text",
                packing=False,
                dataset_batch_size=self.training_args.train_batch_size,
                formatting_func=training_data_formatter_callback.__call__,
                model_init_kwargs=self._model_args if self._model is None else None,
                peft_config=lora_config,
                max_seq_length=max_seq_length,
                preprocess_logits_for_metrics=(lambda _x, _y: fake_logit_tensor), # no need to compute logits and reduce memory usage
                generate_callback=self._generate_completion_callback(training_data_formatter_callback),
                generative_compute_metrics=compute_metrics,
                logger=self.code_logger,
                model_init=self._model_init if self._random_init else None,
            )
        trainer.add_callback(self._log_metric_callback())
        if self.training_args.do_train:
            if self.training_args.resume_from_checkpoint is not None:
                trainer.train(resume_from_checkpoint=self.training_args.resume_from_checkpoint) # Resume from a specific checkpoint
            else:
                # Check if a checkpoint exists
                no_checkpoint = True
                try:
                    trainer.train(resume_from_checkpoint=True)
                except:
                    # Eat the exception
                    no_checkpoint = True
                if no_checkpoint:
                    try:
                        trainer.train()
                    except Exception as e:
                        self.code_logger.error(f"Error in training: {e}")
                        self.code_logger.exception(e)
                        self.code_logger.error("Trying to save the model")
        if self.training_args.do_eval and eval_dataset is not None:
            self.code_logger.info("Running evaluation after training")
            if eval_dataset_is_dict:
                for eval_dataset_name in eval_dataset:
                    self.code_logger.info(f"Evaluating dataset: {eval_dataset_name}")
                    trainer.evaluate(trainer.original_eval_dataset[eval_dataset_name], metric_key_prefix=f"eval_{eval_dataset_name}")
            else:
                trainer.evaluate(trainer.original_eval_dataset)
        # Change the model name and save it
        # load the best model
        try:
            if self.training_args.do_train:
                os.makedirs(self.training_args.output_dir, exist_ok=True)
                new_model_name = os.path.join(self.training_args.output_dir, f"best-model-{time_str}")
                if os.path.exists(new_model_name):
                    shutil.move(new_model_name, f"{new_model_name}-before-{time_str}")
                os.makedirs(new_model_name, exist_ok=True)
            if self.training_args.load_best_model_at_end and trainer.state.best_model_checkpoint is not None and self.training_args.do_train:
                # Copy the best model to a new model
                best_model_path = trainer.state.best_model_checkpoint
                # Recursively copy the best model
                self.code_logger.info(f"Copying the best model from {best_model_path} to {new_model_name}")
                shutil.copytree(best_model_path, new_model_name, dirs_exist_ok=True)    
            else:
                best_model_path = None
        except Exception as e:
            self.code_logger.error(f"Error in copying the best model: {e}")
            best_model_path = None
        if self.training_args.do_train:
            trainer.args.output_dir = new_model_name
            self.code_logger.info(f"Saving the model to {new_model_name}")
            final = os.path.join(self.training_args.output_dir, "final")
            os.makedirs(final, exist_ok=True)
            self.code_logger.info(f"Saving the final model to {final}")
            trainer.save_model(output_dir=final)
        if self._should_use_comet:
            self._comet_experiment.end()

if __name__ == '__main__':
    # Model from Hugging Face hub
    import os
    import numpy as np
    import json
    assert os.path.exists(".secrets/huggingface_token.json"), "Please create a .secrets file with your HF token"
    # model_name = "meta-llama/Llama-2-7b-hf"
    # model_name = "meta-llama/Llama-2-7b-chat-hf"
    model_name = "Salesforce/codet5-small"
    is_seq2seq = "llama-2" not in model_name.lower()
    with open(".secrets/huggingface_token.json", "r") as f:
        token = json.load(f)["token"]
    model = Model(model_name, token=token, use_lora=False, is_seq2seq=is_seq2seq)
    main_prompt = "Do simple math problems (Answer only the number and use '[END]' to finish the response):\nQuestion: 2 + 2\nAnswer: 4\n[END]"
    with model:
        for response in model.generate(
                [
                    f"{main_prompt}\nQuestion: 4 + 5\nAnswer:",
                    f"{main_prompt}\nQuestion: 2 + 2\nAnswer:",
                    f"{main_prompt}\nQuestion: 3 + 3 * 3\nAnswer:",
                ],
                max_new_tokens=10,
                temperature=0.1, # Nucleus sampling
                do_sample=True, # Nucleus sampling
                top_k=5, # Nucleus sampling
                # num_beams=5, # Beam search
                num_return_sequences=5,
                stop_tokens=["[END]"],#, model._tokenizer.eos_token],
                skip_special_tokens=True,
                padding=True,
                #truncation=True,
                compute_probabilities=True,
                return_full_text=False):
            
            print("-" * 50)
            print(f"Prompt: \n{response.input_text}")
            for idx, result in enumerate(response.generated_text):
                probability = np.exp(-response.neg_log_likelihood[idx])
                print(f"Result [{idx + 1}] ({probability*100}% chance): {result}")
            print("-" * 50)
